import numpy as np
from typing import Tuple, Optional
import warnings


def init_library_size(
    data: np.ndarray,
    batch_indices: Optional[np.ndarray] = None,
    n_batch: Optional[int] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """Compute and return library size. It is used to compute the prior on the library size.
    Adapted from https://github.com/scverse/scvi-tools."""
    if batch_indices is None:
        n_batch = 1
        batch_indices = np.zeros(data.shape[0], dtype=np.int32)

    library_log_means = np.zeros(n_batch)
    library_log_vars = np.ones(n_batch)

    for i_batch in np.unique(batch_indices):
        idx_batch = np.squeeze(batch_indices == i_batch)
        batch_data = data[idx_batch.nonzero()
                          [0]]  # h5ad requires integer indexing arrays.
        sum_counts = batch_data.sum(axis=1)
        masked_log_sum = np.ma.log(sum_counts)
        if np.ma.is_masked(masked_log_sum):
            warnings.warn(
                "This dataset has some empty cells, this might fail inference."
                "Data should be filtered with `scanpy.pp.filter_cells()`",
                UserWarning,
            )

        log_counts = masked_log_sum.filled(0)
        library_log_means[i_batch] = np.mean(log_counts).astype(np.float32)
        library_log_vars[i_batch] = np.var(log_counts).astype(np.float32)

    return library_log_means.reshape(1, -1), library_log_vars.reshape(1, -1)



import torch.nn as nn


class scVI(nn.Module):

    def __init__(self):
        import scvi as scvi_package

        super().__init__()
        self.model = None
        self.var_names = None

    def forward(self, adata):
        try:
            self.prepare_data(adata)

        except Exception as error:
            print("No var names found in SCVI reference vars")
            print(f"adata.var.index must include some of {self.var_names}")

            # See docstring for more information. Expect input data to have index
            # ordered and dimensions per scVI specifications
            self.force_data_match(adata)

        # getting variational autoencoder
        vae_q = self.model.load_query_data(adata, self.model)
        vae_q.is_trained = True

        return vae_q.get_latent_representation()

    def load(self):
        import scvi as scvi_package
        import os

        from tdc_ml.multi_pred.anndata_dataset import DataLoader
        from tdc_ml.model_server.model_loaders import scvi_loader

        if not os.path.isdir("scvi_model"):
            loader = scvi_loader.scVILoader()
            loader.load("2024-07-01")

        # adata object needed for loading model
        adata = DataLoader("scvi_test_dataset",
                           "./data",
                           dataset_names=["scvi_test_dataset"],
                           no_convert=True).adata

        # See docstring for more information. Expect index
        # ordered and dimensions per scVI specifications
        self.force_data_match(adata)

        #instantiate SCVI model (not callable, just used to get VAE)
        self.model = scvi_package.model.SCVI.load('scvi_model', adata)
        self.var_names = adata.var.index

        print("loaded scVI:")
        print(f"{self.model}")

        return self.model

    def prepare_data(self, adata):
        import numpy as np

        assert True in np.isin(adata.var.index, self.var_names)

        adata.obs["batch"] = "unassigned"
        self.model.prepare_query_anndata(adata, self.model)

    def force_data_match(self, adata):
        '''
        Input data is expected to have index ordered and dimensions as per scVI specifications.
        For more information visit:
        https://huggingface.co/datasets/scvi-tools/DATASET-FOR-UNIT-TESTING-1/tree/main
        '''
        import torch
        import numpy as np

        metadata = torch.load("scvi_model/model.pt",
                              map_location=torch.device('cpu'))

        # setting indices that match
        adata.var.index = metadata["attr_dict"]["registry_"][
            "field_registries"]["X"]["state_registry"]["column_names"]

        # Padding X so dimensions match. Need 8000 columns because scVI was trained using adata
        # containing 8000 genes. This is the number of var indices extracted from metadata above.
        additional_columns = np.zeros(
            (adata.X.shape[0], 8000 - adata.X.shape[1]))
        adata.X = np.hstack([adata.X, additional_columns])

        # getting a batch name that matches
        adata.obs['batch'] = metadata["attr_dict"]["registry_"][
            "field_registries"]["batch"]["state_registry"][
                "categorical_mapping"][0]



import torch
from torch.nn import ModuleDict
from .nn import Embedding
import torch.nn as nn
from typing import (
    Dict,
    Tuple,
    Union,
)
from abc import abstractmethod
from collections.abc import Iterable
from dataclasses import field, dataclass

LossRecord = Union[Dict[str, torch.Tensor], torch.Tensor]


@dataclass
class LossOutput:
    """Loss signature for models.

    This class provides an organized way to record the model loss, as well as
    the components of the ELBO. This may also be used in MLE, MAP, EM methods.
    The loss is used for backpropagation during inference. The other parameters
    are used for logging/early stopping during inference.

    Parameters
    ----------
    loss
        Tensor with loss for minibatch. Should be one dimensional with one value.
        Note that loss should be in an array/tensor and not a float.
    reconstruction_loss
        Reconstruction loss for each observation in the minibatch. If a tensor, converted to
        a dictionary with key "reconstruction_loss" and value as tensor.
    kl_local
        KL divergence associated with each observation in the minibatch. If a tensor, converted to
        a dictionary with key "kl_local" and value as tensor.
    kl_global
        Global KL divergence term. Should be one dimensional with one value. If a tensor, converted
        to a dictionary with key "kl_global" and value as tensor.
    classification_loss
        Classification loss.
    logits
        Logits for classification.
    true_labels
        True labels for classification.
    extra_metrics
        Additional metrics can be passed as arrays/tensors or dictionaries of
        arrays/tensors.
    n_obs_minibatch
        Number of observations in the minibatch. If None, will be inferred from
        the shape of the reconstruction_loss tensor.


    Examples
    --------
    >>> loss_output = LossOutput(
    ...     loss=loss,
    ...     reconstruction_loss=reconstruction_loss,
    ...     kl_local=kl_local,
    ...     extra_metrics={"x": scalar_tensor_x, "y": scalar_tensor_y},
    ... )
    """

    loss: LossRecord
    reconstruction_loss: LossRecord | None = None
    kl_local: LossRecord | None = None
    kl_global: LossRecord | None = None
    classification_loss: LossRecord | None = None
    logits: torch.Tensor | None = None
    true_labels: torch.Tensor | None = None
    extra_metrics: Dict[str, torch.Tensor] | None = field(default_factory=dict)
    n_obs_minibatch: int | None = None
    reconstruction_loss_sum: torch.Tensor = field(default=None)
    kl_local_sum: torch.Tensor = field(default=None)
    kl_global_sum: torch.Tensor = field(default=None)

    def __post_init__(self):
        object.__setattr__(self, "loss", self.dict_sum(self.loss))

        if self.n_obs_minibatch is None and self.reconstruction_loss is None:
            raise ValueError(
                "Must provide either n_obs_minibatch or reconstruction_loss")

        default = 0 * self.loss
        if self.reconstruction_loss is None:
            object.__setattr__(self, "reconstruction_loss", default)
        if self.kl_local is None:
            object.__setattr__(self, "kl_local", default)
        if self.kl_global is None:
            object.__setattr__(self, "kl_global", default)

        object.__setattr__(self, "reconstruction_loss",
                           self._as_dict("reconstruction_loss"))
        object.__setattr__(self, "kl_local", self._as_dict("kl_local"))
        object.__setattr__(self, "kl_global", self._as_dict("kl_global"))
        object.__setattr__(
            self,
            "reconstruction_loss_sum",
            self.dict_sum(self.reconstruction_loss).sum(),
        )
        object.__setattr__(self, "kl_local_sum",
                           self.dict_sum(self.kl_local).sum())
        object.__setattr__(self, "kl_global_sum", self.dict_sum(self.kl_global))

        if self.reconstruction_loss is not None and self.n_obs_minibatch is None:
            rec_loss = self.reconstruction_loss
            object.__setattr__(self, "n_obs_minibatch",
                               list(rec_loss.values())[0].shape[0])

        if self.classification_loss is not None and (self.logits is None or
                                                     self.true_labels is None):
            raise ValueError(
                "Must provide `logits` and `true_labels` if `classification_loss` is "
                "provided.")

    @staticmethod
    def dict_sum(dictionary: Dict[str, torch.Tensor] | torch.Tensor):
        """Sum over elements of a dictionary."""
        if isinstance(dictionary, dict):
            return sum(dictionary.values())
        else:
            return dictionary

    @property
    def extra_metrics_keys(self) -> Iterable[str]:
        """Keys for extra metrics."""
        return self.extra_metrics.keys()

    def _as_dict(self, attr_name: str):
        attr = getattr(self, attr_name)
        if isinstance(attr, dict):
            return attr
        else:
            return {attr_name: attr}


def _get_dict_if_none(param):
    param = {} if not isinstance(param, dict) else param

    return param


def _generic_forward(
    module,
    tensors,
    inference_kwargs,
    generative_kwargs,
    loss_kwargs,
    get_inference_input_kwargs,
    get_generative_input_kwargs,
    compute_loss,
):
    """Core of the forward call shared by PyTorch- and Jax-based modules."""
    inference_kwargs = _get_dict_if_none(inference_kwargs)
    generative_kwargs = _get_dict_if_none(generative_kwargs)
    loss_kwargs = _get_dict_if_none(loss_kwargs)
    get_inference_input_kwargs = _get_dict_if_none(get_inference_input_kwargs)
    get_generative_input_kwargs = _get_dict_if_none(get_generative_input_kwargs)

    inference_inputs = module._get_inference_input(tensors,
                                                   **get_inference_input_kwargs)
    inference_outputs = module.inference(**inference_inputs, **inference_kwargs)
    generative_inputs = module._get_generative_input(
        tensors, inference_outputs, **get_generative_input_kwargs)
    generative_outputs = module.generative(**generative_inputs,
                                           **generative_kwargs)
    if compute_loss:
        losses = module.loss(tensors, inference_outputs, generative_outputs,
                             **loss_kwargs)
        return inference_outputs, generative_outputs, losses
    else:
        return inference_outputs, generative_outputs


class EmbeddingModuleMixin:
    """``EXPERIMENTAL`` Mixin class for initializing and using embeddings in a module.

    Notes
    -----
    Lifecycle: experimental in v1.2.
    """

    @property
    def embeddings_dict(self) -> ModuleDict:
        """Dictionary of embeddings."""
        if not hasattr(self, "_embeddings_dict"):
            self._embeddings_dict = ModuleDict()
        return self._embeddings_dict

    def add_embedding(self,
                      key: str,
                      embedding: Embedding,
                      overwrite: bool = False) -> None:
        """Add an embedding to the module."""
        if key in self.embeddings_dict and not overwrite:
            raise KeyError(f"Embedding {key} already exists.")
        self.embeddings_dict[key] = embedding

    def remove_embedding(self, key: str) -> None:
        """Remove an embedding from the module."""
        if key not in self.embeddings_dict:
            raise KeyError(f"Embedding {key} not found.")
        del self.embeddings_dict[key]

    def get_embedding(self, key: str) -> Embedding:
        """Get an embedding from the module."""
        if key not in self.embeddings_dict:
            raise KeyError(f"Embedding {key} not found.")
        return self.embeddings_dict[key]

    def init_embedding(
        self,
        key: str,
        num_embeddings: int,
        embedding_dim: int = 5,
        **kwargs,
    ) -> None:
        """Initialize an embedding in the module."""
        self.add_embedding(key,
                           Embedding(num_embeddings, embedding_dim, **kwargs))

    def compute_embedding(self, key: str,
                          indices: torch.Tensor) -> torch.Tensor:
        """Forward pass for an embedding."""
        indices = indices.flatten() if indices.ndim > 1 else indices
        return self.get_embedding(key)(indices)


class BaseModuleClass(nn.Module):
    """Abstract class for scvi-tools modules.

    Notes
    -----
    See further usage examples in the following tutorials:

    1. :doc:`/tutorials/notebooks/dev/module_user_guide`
    """

    def __init__(self,):
        super().__init__()

    @property
    def device(self):
        device = list({p.device for p in self.parameters()})
        if len(device) > 1:
            raise RuntimeError("Module tensors on multiple devices.")
        return device[0]

    def on_load(self, model):
        """Callback function run in :meth:`~scvi.model.base.BaseModelClass.load`."""

    def forward(
        self,
        tensors,
        get_inference_input_kwargs: dict | None = None,
        get_generative_input_kwargs: dict | None = None,
        inference_kwargs: dict | None = None,
        generative_kwargs: dict | None = None,
        loss_kwargs: dict | None = None,
        compute_loss=True,
    ) -> Tuple[torch.Tensor, torch.Tensor] | Tuple[torch.Tensor, torch.Tensor,
                                                   LossOutput]:
        """Forward pass through the network.

        Parameters
        ----------
        tensors
            tensors to pass through
        get_inference_input_kwargs
            Keyword args for ``_get_inference_input()``
        get_generative_input_kwargs
            Keyword args for ``_get_generative_input()``
        inference_kwargs
            Keyword args for ``inference()``
        generative_kwargs
            Keyword args for ``generative()``
        loss_kwargs
            Keyword args for ``loss()``
        compute_loss
            Whether to compute loss on forward pass. This adds
            another return value.
        """
        return _generic_forward(
            self,
            tensors,
            inference_kwargs,
            generative_kwargs,
            loss_kwargs,
            get_inference_input_kwargs,
            get_generative_input_kwargs,
            compute_loss,
        )

    @abstractmethod
    def _get_inference_input(self, tensors: Dict[str, torch.Tensor], **kwargs):
        """Parse tensors dictionary for inference related values."""

    @abstractmethod
    def _get_generative_input(
        self,
        tensors: Dict[str, torch.Tensor],
        inference_outputs: Dict[str, torch.Tensor],
        **kwargs,
    ):
        """Parse tensors dictionary for generative related values."""

    @abstractmethod
    def inference(
        self,
        *args,
        **kwargs,
    ) -> Dict[str, torch.Tensor | torch.distributions.Distribution]:
        """Run the recognition model.

        In the case of variational inference, this function will perform steps related to
        computing variational distribution parameters. In a VAE, this will involve running
        data through encoder networks.

        This function should return a dictionary with str keys and :class:`~torch.Tensor` values.
        """

    @abstractmethod
    def generative(
            self, *args, **kwargs
    ) -> Dict[str, torch.Tensor | torch.distributions.Distribution]:
        """Run the generative model.

        This function should return the parameters associated with the likelihood of the data.
        This is typically written as :math:`p(x|z)`.

        This function should return a dictionary with str keys and :class:`~torch.Tensor` values.
        """

    @abstractmethod
    def loss(self, *args, **kwargs) -> LossOutput:
        """Compute the loss for a minibatch of data.

        This function uses the outputs of the inference and generative functions to compute
        a loss. This many optionally include other penalty terms, which should be computed here.

        This function should return an object of type :class:`~scvi.module.base.LossOutput`.
        """

    @abstractmethod
    def sample(self, *args, **kwargs):
        """Generate samples from the learned model."""



import torch
from torch.distributions import Distribution, Gamma, constraints
import torch.nn.functional as F
from torch.distributions import Poisson as PoissonTorch
from torch.distributions.constraints import Constraint
from torch.distributions.utils import (
    broadcast_all,
    lazy_property,
    logits_to_probs,
    probs_to_logits,
)
from typing import Tuple
import warnings


class _Optional(Constraint):

    def __init__(self, constraint: Constraint):
        self.constraint = constraint

    def check(self, value: torch.Tensor) -> torch.Tensor:
        if value is None:
            return torch.ones(1, dtype=torch.bool)
        return self.constraint.check(value)

    def __repr__(self) -> str:
        return f"Optional({self.constraint})"


def optional_constraint(constraint: Constraint) -> Constraint:
    """Returns a wrapped constraint that allows optional values."""
    return _Optional(constraint)


def _convert_counts_logits_to_mean_disp(
        total_count: torch.Tensor,
        logits: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    """NB parameterizations conversion.

    Parameters
    ----------
    total_count
        Number of failures until the experiment is stopped.
    logits
        success logits.

    Returns
    -------
    type
        the mean and inverse overdispersion of the NB distribution.

    """
    theta = total_count
    mu = logits.exp() * theta
    return mu, theta


def log_nb_positive(
    x: torch.Tensor,
    mu: torch.Tensor,
    theta: torch.Tensor,
    eps: float = 1e-8,
    log_fn: callable = torch.log,
    lgamma_fn: callable = torch.lgamma,
) -> torch.Tensor:
    """Log likelihood (scalar) of a minibatch according to a nb model.

    Parameters
    ----------
    x
        data
    mu
        mean of the negative binomial (has to be positive support) (shape: minibatch x vars)
    theta
        inverse dispersion parameter (has to be positive support) (shape: minibatch x vars)
    eps
        numerical stability constant
    log_fn
        log function
    lgamma_fn
        log gamma function
    """
    log = log_fn
    lgamma = lgamma_fn
    log_theta_mu_eps = log(theta + mu + eps)
    res = (theta * (log(theta + eps) - log_theta_mu_eps) + x *
           (log(mu + eps) - log_theta_mu_eps) + lgamma(x + theta) -
           lgamma(theta) - lgamma(x + 1))

    return res


def _gamma(theta: torch.Tensor, mu: torch.Tensor) -> Gamma:
    concentration = theta
    rate = theta / mu
    # Important remark: Gamma is parametrized by the rate = 1/scale!
    gamma_d = Gamma(concentration=concentration, rate=rate)
    return gamma_d


def log_zinb_positive(
    x: torch.Tensor,
    mu: torch.Tensor,
    theta: torch.Tensor,
    pi: torch.Tensor,
    eps: float = 1e-8,
) -> torch.Tensor:
    """Log likelihood (scalar) of a minibatch according to a zinb model.

    Parameters
    ----------
    x
        Data
    mu
        mean of the negative binomial (has to be positive support) (shape: minibatch x vars)
    theta
        inverse dispersion parameter (has to be positive support) (shape: minibatch x vars)
    pi
        logit of the dropout parameter (real support) (shape: minibatch x vars)
    eps
        numerical stability constant

    Notes
    -----
    We parametrize the bernoulli using the logits, hence the softplus functions appearing.
    """
    # theta is the dispersion rate. If .ndimension() == 1, it is shared for all cells (regardless
    # of batch or labels)
    if theta.ndimension() == 1:
        theta = theta.view(
            1, theta.size(0))  # In this case, we reshape theta for broadcasting

    # Uses log(sigmoid(x)) = -softplus(-x)
    softplus_pi = F.softplus(-pi)
    log_theta_eps = torch.log(theta + eps)
    log_theta_mu_eps = torch.log(theta + mu + eps)
    pi_theta_log = -pi + theta * (log_theta_eps - log_theta_mu_eps)

    case_zero = F.softplus(pi_theta_log) - softplus_pi
    mul_case_zero = torch.mul((x < eps).type(torch.float32), case_zero)

    case_non_zero = (-softplus_pi + pi_theta_log + x *
                     (torch.log(mu + eps) - log_theta_mu_eps) +
                     torch.lgamma(x + theta) - torch.lgamma(theta) -
                     torch.lgamma(x + 1))
    mul_case_non_zero = torch.mul((x > eps).type(torch.float32), case_non_zero)

    res = mul_case_zero + mul_case_non_zero

    return res


class NegativeBinomial(Distribution):
    r"""Negative binomial distribution.

    One of the following parameterizations must be provided:

    (1), (`total_count`, `probs`) where `total_count` is the number of failures until
    the experiment is stopped and `probs` the success probability. (2), (`mu`, `theta`)
    parameterization, which is the one used by scvi-tools. These parameters respectively
    control the mean and inverse dispersion of the distribution.

    In the (`mu`, `theta`) parameterization, samples from the negative binomial are generated as
    follows:

    1. :math:`w \sim \textrm{Gamma}(\underbrace{\theta}_{\text{shape}},
       \underbrace{\theta/\mu}_{\text{rate}})`
    2. :math:`x \sim \textrm{Poisson}(w)`

    Parameters
    ----------
    total_count
        Number of failures until the experiment is stopped.
    probs
        The success probability.
    mu
        Mean of the distribution.
    theta
        Inverse dispersion.
    scale
        Normalized mean expression of the distribution.
    validate_args
        Raise ValueError if arguments do not match constraints
    """

    arg_constraints = {
        "mu": optional_constraint(constraints.greater_than_eq(0)),
        "theta": optional_constraint(constraints.greater_than_eq(0)),
        "scale": optional_constraint(constraints.greater_than_eq(0)),
    }
    support = constraints.nonnegative_integer

    def __init__(
        self,
        total_count: torch.Tensor | None = None,
        probs: torch.Tensor | None = None,
        logits: torch.Tensor | None = None,
        mu: torch.Tensor | None = None,
        theta: torch.Tensor | None = None,
        scale: torch.Tensor | None = None,
        validate_args: bool = False,
    ):
        self._eps = 1e-8
        if (mu is None) == (total_count is None):
            raise ValueError(
                "Please use one of the two possible parameterizations. Refer to the documentation "
                "for more information.")

        using_param_1 = total_count is not None and (logits is not None or
                                                     probs is not None)
        if using_param_1:
            logits = logits if logits is not None else probs_to_logits(probs)
            total_count = total_count.type_as(logits)
            total_count, logits = broadcast_all(total_count, logits)
            mu, theta = _convert_counts_logits_to_mean_disp(total_count, logits)
        else:
            mu, theta = broadcast_all(mu, theta)
        self.mu = mu
        self.theta = theta
        self.scale = scale
        super().__init__(validate_args=validate_args)

    @property
    def mean(self) -> torch.Tensor:
        return self.mu

    @property
    def variance(self) -> torch.Tensor:
        return self.mean + (self.mean**2) / self.theta

    @torch.inference_mode()
    def sample(
        self,
        sample_shape: torch.Size | tuple | None = None,
    ) -> torch.Tensor:
        """Sample from the distribution."""
        sample_shape = sample_shape or torch.Size()
        gamma_d = self._gamma()
        p_means = gamma_d.sample(sample_shape)

        # Clamping as distributions objects can have buggy behaviors when
        # their parameters are too high
        l_train = torch.clamp(p_means, max=1e8)
        counts = PoissonTorch(
            l_train).sample()  # Shape : (n_samples, n_cells_batch, n_vars)
        return counts

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        if self._validate_args:
            try:
                self._validate_sample(value)
            except ValueError:
                warnings.warn(
                    "The value argument must be within the support of the distribution",
                    UserWarning,
                )

        return log_nb_positive(value,
                               mu=self.mu,
                               theta=self.theta,
                               eps=self._eps)

    def _gamma(self) -> Gamma:
        return _gamma(self.theta, self.mu)

    def __repr__(self) -> str:
        param_names = [
            k for k, _ in self.arg_constraints.items() if k in self.__dict__
        ]
        args_string = ", ".join([
            f"{p}: "
            f"{self.__dict__[p] if self.__dict__[p].numel() == 1 else self.__dict__[p].size()}"
            for p in param_names
            if self.__dict__[p] is not None
        ])
        return self.__class__.__name__ + "(" + args_string + ")"


class ZeroInflatedNegativeBinomial(NegativeBinomial):
    r"""Zero-inflated negative binomial distribution.

    One of the following parameterizations must be provided:

    (1), (`total_count`, `probs`) where `total_count` is the number of failures until
    the experiment is stopped and `probs` the success probability. (2), (`mu`, `theta`)
    parameterization, which is the one used by scvi-tools. These parameters respectively
    control the mean and inverse dispersion of the distribution.

    In the (`mu`, `theta`) parameterization, samples from the negative binomial are generated as
    follows:

    1. :math:`w \sim \textrm{Gamma}(\underbrace{\theta}_{\text{shape}},
       \underbrace{\theta/\mu}_{\text{rate}})`
    2. :math:`x \sim \textrm{Poisson}(w)`

    Parameters
    ----------
    total_count
        Number of failures until the experiment is stopped.
    probs
        The success probability.
    mu
        Mean of the distribution.
    theta
        Inverse dispersion.
    zi_logits
        Logits scale of zero inflation probability.
    scale
        Normalized mean expression of the distribution.
    validate_args
        Raise ValueError if arguments do not match constraints
    """

    arg_constraints = {
        "mu": optional_constraint(constraints.greater_than_eq(0)),
        "theta": optional_constraint(constraints.greater_than_eq(0)),
        "zi_logits": optional_constraint(constraints.real),
        "scale": optional_constraint(constraints.greater_than_eq(0)),
    }
    support = constraints.nonnegative_integer

    def __init__(
        self,
        total_count: torch.Tensor | None = None,
        probs: torch.Tensor | None = None,
        logits: torch.Tensor | None = None,
        mu: torch.Tensor | None = None,
        theta: torch.Tensor | None = None,
        zi_logits: torch.Tensor | None = None,
        scale: torch.Tensor | None = None,
        validate_args: bool = False,
    ):
        super().__init__(
            total_count=total_count,
            probs=probs,
            logits=logits,
            mu=mu,
            theta=theta,
            scale=scale,
            validate_args=validate_args,
        )
        self.zi_logits, self.mu, self.theta = broadcast_all(
            zi_logits, self.mu, self.theta)

    @property
    def mean(self) -> torch.Tensor:
        pi = self.zi_probs
        return (1 - pi) * self.mu

    @property
    def variance(self) -> None:
        raise NotImplementedError

    @lazy_property
    def zi_logits(self) -> torch.Tensor:
        """ZI logits."""
        return probs_to_logits(self.zi_probs, is_binary=True)

    @lazy_property
    def zi_probs(self) -> torch.Tensor:
        return logits_to_probs(self.zi_logits, is_binary=True)

    @torch.inference_mode()
    def sample(
        self,
        sample_shape: torch.Size | tuple | None = None,
    ) -> torch.Tensor:
        """Sample from the distribution."""
        sample_shape = sample_shape or torch.Size()
        samp = super().sample(sample_shape=sample_shape)
        is_zero = torch.rand_like(samp) <= self.zi_probs
        samp_ = torch.where(is_zero, torch.zeros_like(samp), samp)
        return samp_

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        """Log probability."""
        try:
            self._validate_sample(value)
        except ValueError:
            warnings.warn(
                "The value argument must be within the support of the distribution",
                UserWarning,
            )
        return log_zinb_positive(value,
                                 self.mu,
                                 self.theta,
                                 self.zi_logits,
                                 eps=1e-08)


class Poisson(PoissonTorch):
    """Poisson distribution.

    Parameters
    ----------
    rate
        rate of the Poisson distribution.
    validate_args
        whether to validate input.
    scale
        Normalized mean expression of the distribution.
        This optional parameter is not used in any computations, but allows to store
        normalization expression levels.

    """

    def __init__(
        self,
        rate: torch.Tensor,
        validate_args: bool | None = None,
        scale: torch.Tensor = None,
    ):
        super().__init__(rate=rate, validate_args=validate_args)
        self.scale = scale

    def __repr__(self) -> str:
        param_names = [
            k for k, _ in self.arg_constraints.items() if k in self.__dict__
        ]
        args_string = ", ".join([
            f"{p}: "
            f"{self.__dict__[p] if self.__dict__[p].numel() == 1 else self.__dict__[p].size()}"
            for p in param_names
            if self.__dict__[p] is not None
        ])
        return self.__class__.__name__ + "(" + args_string + ")"



import torch.nn as nn
import torch
import torch.nn.functional as F
from torch.distributions import Normal, Distribution
import numpy as np
from typing import (
    Dict,
    Optional,
    Union,
    Tuple,
    Any,
    Callable,
    List,
)
from ..cvae.module import CVAE
from ..cvae.base import LossOutput
from ..cvae.utils import init_library_size
from . import prepare_cell_text_llm
from transformers import (
    PreTrainedModel,
    PretrainedConfig,
    AutoConfig,
    AutoTokenizer,
    AutoModelForSeq2SeqLM,
    AutoModelForCausalLM,
    CONFIG_MAPPING,
    PreTrainedTokenizer,
    PreTrainedTokenizerFast,
    Blip2QFormerConfig,
    Blip2QFormerModel,
)
from transformers.generation.streamers import BaseStreamer
from transformers.modeling_outputs import (
    Seq2SeqLMOutput,
    CausalLMOutput,
)

ACTIVATION_FAMILY = {
    "gelu": F.gelu,
    "relu": F.relu,
    "tanh": F.tanh,
    "sigmoid": F.sigmoid,
    "selu": F.selu,
    "silu": F.silu,
}


def get_torch_activation(name: str) -> Callable[[torch.Tensor], torch.Tensor]:
    """Get the torch activation function by name (e.g, ``get_torch_activation("gelu")``)."""
    func = ACTIVATION_FAMILY.get(name, None)
    if callable(func):
        return func

    raise ValueError(f"Unknown activation function: {name}.")


class ResidualBlock(nn.Module):
    """
    A PyTorch module implementing a residual block. It is composed of a linear layer, an activation
    function, and a dropout layer.

    Parameters
    ----------
    n_inputs: int
        The number of input features (input dimension) for the linear layer.
    n_outputs: int
        The number of output features (output dimension) for the linear layer. If a residual
        connection is enabled, ``n_inputs`` must be equal to ``n_outputs``.
    activation_func: str, default "gelu"
        The name of the activation function to apply after the linear transformation. It should
        be one of "gelu", "relu", "tanh", "sigmoid", "selu", or "silu".
    dropout_rate: float, default 0.1
        The probability of an element to be zeroed in the dropout layer applied after the
        activation function.
    residual: bool, default True
        If True, the input will be added to the output (residual connection), assuming the
        input and output dimensions are the same. If False, no residual connection is used.
    """

    def __init__(
        self,
        n_inputs: int,
        n_outputs: int,
        activation_func: str = "gelu",
        dropout_rate: float = 0.1,
        residual: bool = True,
    ) -> "ResidualBlock":
        super(ResidualBlock, self).__init__()
        if residual:
            assert n_inputs == n_outputs, "Residual connection requires the same input and output dimensions"

        self.residual = residual
        self.linear = nn.Linear(n_inputs, n_outputs)
        self.activation = get_torch_activation(activation_func)
        self.dropout = nn.Dropout(dropout_rate)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Given the input ``x``, get the corresponding output of the residual block."""
        if self.residual:
            return x + self.dropout(self.activation(self.linear(x)))
        else:
            return self.dropout(self.activation(self.linear(x)))


class SCQFormer(Blip2QFormerModel):
    """
    A variant of the BLIP2 Q-Former model designed for encoding cell counts.

    The SCQFormer takes in a set of input cell counts, passes them through a series of residual blocks
    (cell encoders) to generate key-value pairs. Q-Former is used to pool the key-value pairs via query
    tokens to generate the final output.

    Parameters
    ----------
    count_dim: int
        The dimensionality of the input cell counts. This corresponds to the number of input features.
    num_query_tokens: int
        The number of query tokens generated by the model. These tokens interact with the hidden states
        from residual blocks through cross-attention layers and interact with each other through
        self-attention layers.
    num_key_value_tokens: int
        The number of key-value tokens used in the cross-attention mechanism.
    config: transformers.Blip2QFormerConfig
        The configuration object that contains hyperparameters for the ``transformers.Blip2QFormer`` model.
    num_hidden_layers: int, default 2
        The number of ``ResidualBlock`` in the cell encoder.
    """

    def __init__(
        self,
        count_dim: int,
        num_query_tokens: int,
        num_key_value_tokens: int,
        config: Blip2QFormerConfig,
        num_hidden_layers: int = 2,
    ) -> "SCQFormer":
        # initialize the Blip2QFormerModel and model weights
        super().__init__(config)

        hidden_dropout_prob = config.hidden_dropout_prob
        hidden_act = config.hidden_act
        input_dim, output_dim = count_dim, num_key_value_tokens * config.hidden_size

        self.cell_encoder = []
        for i in range(num_hidden_layers):
            if i == 0:
                self.cell_encoder.append(
                    ResidualBlock(
                        input_dim,
                        output_dim,
                        activation_func=hidden_act,
                        dropout_rate=hidden_dropout_prob,
                        residual=False,
                    ))
            elif i == num_hidden_layers - 1:
                self.cell_encoder.append(
                    ResidualBlock(
                        output_dim,
                        output_dim,
                        activation_func=hidden_act,
                        dropout_rate=0.0,
                        residual=True,
                    ))
            else:
                self.cell_encoder.append(
                    ResidualBlock(
                        output_dim,
                        output_dim,
                        activation_func=hidden_act,
                        dropout_rate=hidden_dropout_prob,
                        residual=True,
                    ))
        self.cell_encoder = nn.Sequential(*self.cell_encoder)
        self.query_tokens = nn.Parameter(
            torch.zeros(1, num_query_tokens, config.hidden_size))

        # initialize the weights of the model again
        self.init_weights()
        nn.init.normal_(self.query_tokens, std=config.initializer_range)

    def forward(self, input_counts: torch.Tensor) -> torch.Tensor:
        """Encode the input cell counts and generate the final output."""
        encoder_hidden_states = self.cell_encoder(input_counts)
        batch_size = encoder_hidden_states.size(0)
        encoder_hidden_states = torch.reshape(
            encoder_hidden_states, (batch_size, -1, self.config.hidden_size))

        query_tokens = self.query_tokens.expand(batch_size, -1, -1)
        outputs = super().forward(
            query_tokens, encoder_hidden_states=encoder_hidden_states)[0]

        return outputs


class Generator(nn.Module):
    """
    A Generator model for conditional generation of gene expression profiles based on
    conditional embeddings and latent variables. This model is built upon a Conditional
    Variational Autoencoder (CVAE), which integrates continuous and categorical covariates
    to generate synthetic data.

    Parameters
    ----------
    n_input: int
        The dimensionality of the input data (i.e., the number of genes).
    condition_dim: int, default 4096
        The dimensionality of the condition embedding input, which represents the condition
        that controls the generative process (e.g., cell type embeddings).
    condition_input_dim: int, default 128
        The dimensionality of the mapped condition after processing by the mapping layer.
    **kwargs: dict, optional
        Additional arguments passed to the base model, typically the ``scvi.module.CVAE`` model.
    """

    def __init__(self,
                 n_input: int,
                 condition_dim: int = 4096,
                 condition_input_dim: int = 128,
                 **kwargs) -> "Generator":

        super(Generator, self).__init__()
        self.mapping_layer = nn.Linear(condition_dim, condition_input_dim)
        kwargs["n_continuous_cov"] = condition_input_dim
        self.base_model = CVAE(n_input, **kwargs)

    def forward(
        self,
        condition_embeds: torch.Tensor,
        gene_counts: torch.Tensor,
        batch_ids: Optional[torch.Tensor] = None,
        cat_covs: Optional[torch.Tensor] = None,
        labels: Optional[torch.Tensor] = None,
        loss_kwargs: Optional[Dict[str, Any]] = None,
        compute_loss: bool = True,
    ) -> Union[Tuple[Dict[str, torch.Tensor | Distribution | None], Dict[
            str, Distribution | None]], LossOutput]:
        """
        Perform a forward pass through the generator.

        Parameters
        ----------
        condition_embeds: torch.Tensor
            The condition embeddings control the generation process.
        gene_counts: torch.Tensor
            A tensor representing observed gene counts, which serves as input for training.
        batch_ids: torch.Tensor, optional, default None
            The batch identifiers for each sample in the input data. If None, all samples are assigned
            to batch 0. Another way to consider batch information is to use  ``condition_embeds`` to encode
            batch information.
        cat_covs: torch.Tensor, optional, default None
            Categorical covariates related to the sample.
        labels: torch.Tensor, optional, default None
            Class labels associated with the samples.
        loss_kwargs: dict, optional, default None
            Additional keyword arguments passed to the loss function ``scvi.module.CVAE.loss`` (e.g., the weight of
            the KL divergence term between latent variables and the prior distribution).
        compute_loss: bool, default True
            Whether to compute and return the loss during the forward pass. If False, the outputs of
            inference process and generative process are returned. If True, the loss is returned.

        Returns
        -------
        outputs: Union[Tuple[Dict[str, torch.Tensor | Distribution | None], Dict[str, Distribution | None]], LossOutput]
            The outputs of the model. If ``compute_loss`` is True, the loss is returned. Otherwise, the outputs
            of the inference process and generative process are returned.
        """
        if condition_embeds.ndim > 2:
            condition_embeds = condition_embeds.squeeze(1)
        if batch_ids is None:
            batch_ids = torch.zeros(
                (gene_counts.shape[0], 1)).to(dtype=torch.long,
                                              device=gene_counts.device)
        inputs = {
            "gene_counts": gene_counts,
            "batch_ids": batch_ids,
            "cat_covs": cat_covs,
            "y": labels,
        }
        cont_covs = self.mapping_layer(condition_embeds)
        inputs["cont_covs"] = cont_covs

        outputs = self.base_model(inputs,
                                  loss_kwargs=loss_kwargs,
                                  compute_loss=compute_loss)
        if compute_loss:
            return outputs[-1]

        return outputs

    @torch.inference_mode()
    def generate(
        self,
        condition_embeds: torch.Tensor,
        batch_ids: Optional[torch.Tensor] = None,
        library_size: Optional[torch.Tensor] = None,
        batch_size: int = 128,
        to_numpy: bool = True,
    ) -> torch.Tensor | np.ndarray:
        """
        Generate synthetic gene expression profiles from the latent space.

        Parameters
        ----------
        condition_embeds: torch.Tensor
            The condition embeddings used to control the generation process.
        batch_ids: torch.Tensor, optional, default None
            The batch identifiers for each sample in the input data. If None, all samples are assigned
            to batch 0. Another way to consider batch information is to use  ``condition_embeds`` to encode
            batch information.
        library_size: torch.Tensor, optional, default None
            The library size vector predefined for each sample. It is used to control the library size of each
            generated cell. If None, a prior distribution over library size will be used to sample the library
            size for each cell. Note that, in general, the prior distribution is set by
            ``scvi.utils.init_library_size``.
        batch_size: int, default 128
            The batch size for generating data. Control the number of samples processed at once.
        to_numpy: bool, default True
            Whether to return the output as a NumPy array. If False, a PyTorch tensor is returned.

        Returns
        -------
        outputs: torch.Tensor | np.ndarray
            The generated gene expression data. The shape of the output depends on the input shape
            and batch size. If ``to_numpy`` is True, the output is a NumPy array, otherwise it is a
            PyTorch tensor.
        """
        if condition_embeds.ndim > 2:
            condition_embeds = condition_embeds.squeeze(1)
        cont_covs = self.mapping_layer(condition_embeds)
        n_samples = condition_embeds.shape[0]
        if batch_ids is None:
            batch_ids = torch.zeros(
                (n_samples, 1)).to(dtype=torch.long,
                                   device=condition_embeds.device)
        device = next(self.base_model.parameters()).device

        outputs = []
        for i in range(0, n_samples, batch_size):
            inputs = {}
            cont_covs_batch = cont_covs[
                i:i + batch_size] if cont_covs is not None else None
            batch_ids_batch = batch_ids[i:i + batch_size]
            # the prior distribution of latent vectors
            p_z = Normal(
                torch.zeros(
                    (batch_ids_batch.size(0), self.base_model.n_latent)),
                torch.ones((batch_ids_batch.size(0), self.base_model.n_latent)))
            inputs = {
                "z": p_z.sample(),
                "batch_ids": batch_ids_batch,
                "cont_covs": cont_covs_batch,
            }
            if library_size is not None:
                inputs["library"] = library_size[i:i + batch_size]
            else:
                p_library = self.base_model.get_prior_library_distribution(
                    batch_ids_batch)
                assert p_library is not None, "The model does not have a prior distribution for library size, please provide it"
                inputs["library"] = p_library.sample()
            inputs = {k: v.to(device) for k, v in inputs.items()}
            p_gene = self.base_model.generative(**inputs)["p_gene"]
            fake_samples = p_gene.sample()
            if fake_samples.device.type == "cuda":
                fake_samples = fake_samples.cpu()
            outputs.append(fake_samples)
        outputs = torch.concat(outputs,
                               dim=0) if len(outputs) > 1 else outputs[0]

        return outputs if not to_numpy else outputs.numpy()


class CellTextLLM(nn.Module):
    """
    A model class that integrates cell encoders, language models, and feature decoders.

    The model can merge encoded cell features with tokenized input text, generating both text and cell data.

    Parameters
    ----------
    base_model: transformers.PreTrainedModel or torch.nn.Module
        The underlying language model that will be used for text generation.
    tokenizer: transformers.PreTrainedTokenizer or transformers.PreTrainedTokenizerFast
        The tokenizer to tokenize the input text sequences.
    feature_encoder: transformers.PreTrainedModel or torch.nn.Module or None, optional, default None
        A model used to encode input cell features, such as cell counts.
    feature_decoder: transformers.PreTrainedModel or torhc.nn.Module or None, optional, default None
        A model used to decode cell features from the hidden states of the language model.
    normalize_total: bool, default True
        Whether to normalize the total input cell counts by dividing by the total sum of the counts.
    log_variational: bool, default True
        Whether to apply a log transformation (log1p) to the input cell counts before encoding. If
        both normalization and log transformation are enabled, the log transformation is applied after
        normalization.
    """

    def __init__(
        self,
        base_model: PreTrainedModel | nn.Module,
        tokenizer: PreTrainedTokenizer | PreTrainedTokenizerFast,
        feature_encoder: PreTrainedModel | nn.Module | None = None,
        feature_decoder: PreTrainedModel | nn.Module | None = None,
        normalize_total: bool = True,
        log_variational: bool = True,
    ) -> "CellTextLLM":
        super(CellTextLLM, self).__init__()
        self.base_model = base_model
        self.tokenizer = tokenizer
        if feature_encoder is not None:
            self.feature_encoder = feature_encoder
        else:
            self.register_buffer("feature_encoder", None)
        if feature_decoder is not None:
            self.feature_decoder = feature_decoder
        else:
            self.register_buffer("feature_decoder", None)

        assert hasattr(self.base_model, "config"),  \
            "Base model must have a config attribute. " + \
            "Please use prepare_cell_text_llm to prepare the model."
        assert hasattr(self.base_model.config, "eos_token_id"), \
            "Base model config must have eos_token_id attribute. " + \
            "Please use prepare_cell_text_llm to prepare the model."

        if self.feature_encoder is not None or self.feature_decoder is not None:
            assert hasattr(self.base_model.config, "special_tokens_index_dict"), \
                "Base model config must have special_tokens_index_dict attribute. " + \
                "Please use prepare_cell_text_llm to prepare the model."
            if self.feature_encoder is not None:
                assert hasattr(self.base_model.config, "pad_token_id"), \
                    "Base model config must have pad_token_id attribute. " + \
                    "Please use prepare_cell_text_llm to prepare the model."
                assert hasattr(self.base_model.config, "ignore_index"), \
                    "Base model config must have ignore_index attribute. " + \
                    "Please use prepare_cell_text_llm to prepare the model."
            if self.feature_decoder is not None:
                assert hasattr(self.base_model.config, "num_signal_tokens"), \
                    "Base model config must have num_signal_tokens attribute. " + \
                    "Please use prepare_cell_text_llm to prepare the model."

        self.normalize_total = normalize_total
        self.log_variational = log_variational

    def _merge_input_ids_with_features(
        self,
        input_ids: torch.Tensor,
        inputs_embeds: torch.Tensor,
        features: torch.Tensor,
        attention_mask: torch.Tensor,
        labels: Optional[torch.Tensor] = None,
    ) -> Dict[str, torch.Tensor | Dict[str, torch.Tensor] | None]:
        """
        Merge input token embeddings with encoded cell features and updates attention masks and labels accordingly.

        This method is designed to handle a mix of text and cell feature tokens. It identifies placeholder tokens in the
        input text sequence (denoted by special tokens) and replaces them with corresponding cell features. It then constructs
        the final sequence of embeddings that contains both text and feature embeddings, ensuring that the attention masks and
        labels are updated accordingly for a consistent model input.

        Adapted from https://github.com/huggingface/transformers/blob/v4.40.1/src/transformers/models/llava/modeling_llava.py#L355.

        Parameters
        ----------
        input_ids: torch.Tensor
            The input token IDs of shape (batch_size, sequence_length), where some tokens may be special placeholders
            indicating the positions to insert cell features.
        inputs_embeds: torch.Tensor
            Precomputed embeddings corresponding to the ``input_ids`` of shape (batch_size, sequence_length, embed_dim).
        features: torch.Tensor
            Encoded cell features of shape (num_cells, num_features, embed_dim). These features will be inserted into
            ``input_embeds`` based on the placeholders in ``input_ids``. If the features are 2D, they will be unsqueezed
            to 3D with an additional dimension.
        attention_mask: torch.Tensor
            Attention mask indicating which tokens should be attended to, of shape (batch_size, sequence_length).
        labels: torch.Tensor, optional, default None
            Ground truth labels for token prediction tasks of shape (batch_size, sequence_length). These labels will also
            be updated to match the new token positions after merging.

        Returns
        -------
        outputs: dict
            A dict containing:
            - "inputs_embeds": torch.Tensor
              The final merged embeddings with both text and cell feature embeddings of shape (batch_size, max_sequence_length, embed_dim).
            - "attention_mask": torch.Tensor
              The updated attention mask reflecting the new merged token sequence of shape (batch_size, max_sequence_length).
            - "labels": torch.Tensor, optional
              The updated labels of shape (batch_size, max_sequence_length), None if labels are not provided.

        Notes
        -----
        - This function assumes that the placeholders for cell features in the ``input_ids`` are identified by a special token.
          The exact special token index is retrieved from the ``base_model.config.special_tokens_index_dict``. The final merged
          sequence is padded to the maximum sequence length derived from the maximum number of input embeddings in the batch.
        - The number of placeholders in the input text sequence should match the number of cell features provided.
        - For encoder-decoder models, the labels are not updated because the labels do not contain input tokens.
        """
        # both MLP-based and transformer-based encoders are considered
        if len(features.shape) == 2:
            features = features.unsqueeze(dim=1)

        num_features, num_feature_legnth, embed_dim = features.shape
        placeholder_index = self.base_model.config.special_tokens_index_dict[
            "placeholder"]
        batch_size, sequence_length = input_ids.shape
        left_padding = not torch.sum(
            input_ids[:,
                      -1] == torch.tensor(self.base_model.config.pad_token_id))

        # 1. create a mask to know where special placeholders are
        special_token_mask = input_ids == placeholder_index
        # for each sample, we calculate the number of placeholders for this modality
        num_special_tokens = torch.sum(special_token_mask, dim=-1)
        # compute the maximum embed dimension
        # if the size of feature is equal to 1, the sequence length after merging is the same
        # we just replace the placeholder with the corresponding features
        max_embed_dim = (num_special_tokens.max() *
                         (num_feature_legnth - 1)) + sequence_length
        batch_indices, non_modality_indices = torch.where(
            input_ids != placeholder_index)

        # 2. compute the positions where text should be written
        # calculate new positions for text tokens in merged modality-text sequence
        # `special_token_mask` identifies placeholders
        # each placeholder will be replaced by `nb_text_tokens_per_feature` text tokens
        # `torch.cumsum` computes how each modality token shifts subsequent text token positions
        # - 1 to adjust for zero-based indexing, as `cumsum` inherently increases indices by one
        new_token_positions = torch.cumsum(
            (special_token_mask * (num_feature_legnth - 1) + 1), -1) - 1
        nb_pad = max_embed_dim - 1 - new_token_positions[:, -1]
        if left_padding:
            new_token_positions += nb_pad[:, None]  # offset for left padding
        text_to_overwrite = new_token_positions[batch_indices,
                                                non_modality_indices]

        # 3. create the full embedding, already padded to the maximum position
        final_embedding = torch.zeros(batch_size,
                                      max_embed_dim,
                                      embed_dim,
                                      dtype=inputs_embeds.dtype,
                                      device=inputs_embeds.device)
        final_attention_mask = torch.zeros(batch_size,
                                           max_embed_dim,
                                           dtype=attention_mask.dtype,
                                           device=inputs_embeds.device)
        if labels is not None and self.base_model.config.is_decoder:
            final_labels = torch.full(
                (batch_size, max_embed_dim),
                self.base_model.config.ignore_index,
                dtype=input_ids.dtype,
                device=input_ids.device,
            )
        # in case the modality encoder or the Language model has been offloaded to CPU, we need to manually
        # set the corresponding tensors into their correct target device
        target_device = inputs_embeds.device
        batch_indices, non_modality_indices, text_to_overwrite = (
            batch_indices.to(target_device),
            non_modality_indices.to(target_device),
            text_to_overwrite.to(target_device),
        )
        attention_mask = attention_mask.to(target_device)

        # 4. fill the embeddings based on the mask
        # if we have ["hey" "<CELL>", "how", "are"]
        # we need to index copy on [0, 577, 578, 579] for the text and [1:576] for the cell features
        final_embedding[batch_indices,
                        text_to_overwrite] = inputs_embeds[batch_indices,
                                                           non_modality_indices]
        final_attention_mask[batch_indices, text_to_overwrite] = attention_mask[
            batch_indices, non_modality_indices]
        if labels is not None and self.base_model.config.is_decoder:
            final_labels[batch_indices,
                         text_to_overwrite] = labels[batch_indices,
                                                     non_modality_indices]

        # 5. fill the embeddings corresponding to the modality. Anything that is still zeros needs filling
        # B * L
        modality_to_overwrite = torch.all(final_embedding == 0, dim=-1)
        modality_to_overwrite &= modality_to_overwrite.cumsum(
            -1) - 1 >= nb_pad[:, None].to(target_device)
        if modality_to_overwrite.sum() != features.shape[:-1].numel():
            raise ValueError(
                f"The input provided to the model are wrong. The number of cell tokens is {torch.sum(special_token_mask)} while"
                f" the number of cells given to the model is {num_features}. This prevents correct indexing and breaks batch generation."
            )
        final_embedding[modality_to_overwrite] = features.contiguous().reshape(
            -1, embed_dim).to(target_device)
        final_attention_mask |= modality_to_overwrite

        # 6. mask out the embedding at padding positions, as we later use the past_key_value value to determine the non-attended tokens
        batch_indices, pad_indices = torch.where(
            input_ids == self.base_model.config.pad_token_id)
        indices_to_mask = new_token_positions[batch_indices, pad_indices]
        final_embedding[batch_indices, indices_to_mask] = 0
        if labels is None or self.base_model.config.is_encoder_decoder:
            # for encoder-decoder models, we don't need to update the labels
            # because the labels don't contain input tokens
            final_labels = None
        labels = final_labels
        attention_mask = final_attention_mask
        inputs_embeds = final_embedding

        return {
            "inputs_embeds": inputs_embeds,
            "attention_mask": attention_mask,
            "labels": labels,
        }

    def _get_condition_embeds(
        self,
        last_hidden_states: torch.Tensor,
        input_ids: torch.Tensor,
        output_counts: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """
        Extract conditional embeddings from the hidden states of the language model.

        This function is used to retrieve the embeddings corresponding to a set of signal tokens
        in the input sequence. These signal tokens are used to indicate the presence of specific conditional information
        (e.g., cell-related information). The embeddings of these tokens are extracted and used to generate gene profiles.

        Parameters
        ----------
        last_hidden_states: torch.Tensor
            The last hidden states of the language model's forward pass, of shape
            (batch_size, sequence_length, hidden_size).
        input_ids: torch.Tensor
            The input token IDs of shape (batch_size, sequence_length), which include the special
            tokens used to identify the locations of the conditional information.
        output_counts: torch.Tensor, optional, default None
            If provided, this tensor contains the expected output cell counts. It is used to verify
            that the number of samples containing signal tokens matches the number of expected outputs.

        Returns
        -------
        condition_embeds: torch.Tensor
            A tensor containing the extracted conditional embeddings from the hidden states,
            of shape (num_cells, num_signal_tokens, hidden_size). These embeddings can be used
            for generating gene profiles.

        Notes
        -----
        - This function assumes that the signal tokens in an input sequence are consecutive.
        - The number of samples containing signal tokens must match the number of expected conditional outputs
        (e.g., gene counts) when provided.
        """
        # get conditional embeddings
        num_signal_tokens = self.base_model.config.num_signal_tokens
        first_signal_token_index = self.base_model.config.special_tokens_index_dict[
            "first_signal_token"]
        output_ids = input_ids
        batch_indices, signal_token_indices = torch.where(
            output_ids == first_signal_token_index)
        if output_counts is not None and len(batch_indices) != len(
                output_counts):
            raise ValueError(
                f"The number of first signal tokens in inputs is {len(batch_indices)} "
                f"but the number of output counts you provide is {len(output_counts)}. "
                f"The row index of each signal token listed as follows:\n{batch_indices}"
            )
        indexer = signal_token_indices[:, None] + torch.arange(
            num_signal_tokens)[None].to(device=output_ids.device,
                                        dtype=output_ids.dtype)
        target_ids = torch.arange(
            first_signal_token_index, first_signal_token_index +
            num_signal_tokens).to(device=output_ids.device,
                                  dtype=output_ids.dtype)
        if not torch.all(
                torch.gather(output_ids[batch_indices], 1, indexer) ==
                target_ids[None]):
            raise ValueError(
                "The arrangement of signal tokens is not consecutive.")
        batch_indices = batch_indices[:, None].expand_as(indexer)
        condition_embeds = last_hidden_states[batch_indices, indexer]

        return condition_embeds

    @torch.inference_mode()
    def generate(
        self,
        input_ids: torch.Tensor,
        input_counts: Optional[torch.Tensor] = None,
        return_conditions: bool = False,
        attention_mask: Optional[torch.Tensor] = None,
        prefix_allowed_tokens_fn: Optional[Callable[[int, torch.Tensor],
                                                    List[int]]] = None,
        synced_gpus: Optional[bool] = None,
        assistant_model: Optional["PreTrainedModel"] = None,
        streamer: Optional["BaseStreamer"] = None,
        **kwargs,
    ) -> Dict[str, List[str | None | np.ndarray]]:
        """
        Generate texts or cells from text sequences ``input_ids`` and input counts ``input_counts`` (optionally).

        This method combines tokenized text input with encoded cell features to generate both text
        sequences and optionally cell data. It handles the integration of text and cell data using
        a feature encoder and decoder if they are provided.

        Parameters
        ----------
        input_ids: torch.Tensor
            Input token IDs of shape (batch_size, sequence_length) for the language model.
        input_counts: torch.Tensor, optional, default None
            Input counts of shape (num_cells, count_dim), which are passed through the feature encoder.
            If not provided, text-only generation is performed.
        return_conditions: bool, optional, default False
            Whether to return the condition embeddings (i.e. the last hidden states of signal tokens).
        attention_mask: torch.Tensor, optional, default None
            A mask that specifies which tokens should be attended to. If not provided, it is computed from ``input_ids``.
        prefix_allowed_tokens_fn: Callable[[int, torch.Tensor], List[int]], optional, default None
            If provided, this function constraints the beam search to allowed tokens only at each step. If not provided,
            no constraint is applied. Please see
            https://huggingface.co/docs/transformers/v4.44.2/en/main_classes/text_generation#transformers.GenerationMixin.generate
            for more details.
        synced_gpus: bool, optional, default None
            Whether to continue running the while loop until max_length. Please see
            https://huggingface.co/docs/transformers/v4.44.2/en/main_classes/text_generation#transformers.GenerationMixin.generate
            for more details.
        assistant_model: PreTrainedModel, optional, default None
            An assistant model that can be used to accelerate generation. The assistant model must have the exact same tokenizer.
            The acceleration is achieved when forecasting candidate tokens with the assistent model is much faster than running
            generation with the model you are calling generate from. As such, the assistant model should be much smaller.
            For more details, please see https://arxiv.org/abs/2302.01318.
        streamer: BaseStreamer, optional, default None
            Streamer object that will be used to stream the generated sequences. Generated tokens are passed through
            ``streamer.put(token_ids)`` and the streamer is responsible for any further processing.
        **kwargs
            Additional keyword arguments passed to the ``generate`` method of the language model. In general,
            the language model is based on the transformers library. Thus, please see
            https://huggingface.co/docs/transformers/v4.44.2/en/main_classes/text_generation#transformers.GenerationMixin.generate
            for more details.

        Returns
        -------
        outputs: dict
            A dict containing the following keys:
            - "texts": list of str, the generated text sequences. Special tokens are removed.
            - "cells": list of cell data, where each entry is the decoded cell features. If no signal tokens are detected during
            generation, each entry is None.
            - "conditions": list of numpy array, where each entry is the condition embeddings. If no signal tokens are detected during
            generation, each entry is None. This key is only present if ``return_conditions`` is True.
        """
        # we don't need custom stopping criteria because the built-in stopping criteria is good enough
        # we will use EosTokenCriteria by setting generation configuration's eos token id
        # to the model's eos token id and the first signal token id
        eos_token_id = [self.base_model.config.eos_token_id]
        if self.feature_decoder is not None:
            eos_token_id.append(self.base_model.config.
                                special_tokens_index_dict["first_signal_token"])

        if attention_mask is None:
            attention_mask = (input_ids
                              != self.base_model.config.pad_token_id).to(
                                  input_ids.device)
            attention_mask = attention_mask.long()
        inputs_embeds = self.base_model.get_input_embeddings()(input_ids)
        if input_counts is not None:
            if self.normalize_total:
                input_counts = input_counts / input_counts.sum(
                    dim=-1, keepdim=True) * 1e4
            if self.log_variational:
                input_counts = torch.log1p(input_counts)
            assert self.feature_encoder is not None, "Feature encoder is not provided."
            features = self.feature_encoder(input_counts)
            merged_inputs = self._merge_input_ids_with_features(
                input_ids,
                inputs_embeds,
                features,
                attention_mask,
            )
            inputs_embeds = merged_inputs["inputs_embeds"]
            attention_mask = merged_inputs["attention_mask"]

        generated_ids = self.base_model.generate(
            inputs_embeds=inputs_embeds,
            attention_mask=attention_mask,
            prefix_allowed_tokens_fn=prefix_allowed_tokens_fn,
            synced_gpus=synced_gpus,
            assistant_model=assistant_model,
            streamer=streamer,
            eos_token_id=eos_token_id,
            **kwargs,
        )
        outputs = {"cells": [None] * generated_ids.size(0)}
        if return_conditions:
            outputs["conditions"] = [None] * generated_ids.size(0)

        if self.feature_decoder is not None:
            # start to generate the cell
            num_signal_tokens = self.base_model.config.num_signal_tokens
            first_signal_token_index = self.base_model.config.special_tokens_index_dict[
                "first_signal_token"]
            batch_size, length = generated_ids.shape
            pad_token_id = self.base_model.config.pad_token_id
            cell_indices, signal_token_indices = torch.where(
                generated_ids == first_signal_token_index)
            if len(cell_indices) > 0:
                nb_pad = max(
                    (signal_token_indices + num_signal_tokens - 1).max().item()
                    - length + 1, 0)
                if nb_pad > 0:
                    final_generated_ids = torch.concat([
                        generated_ids,
                        torch.full((batch_size, nb_pad), pad_token_id).to(
                            device=generated_ids.device,
                            dtype=generated_ids.dtype)
                    ],
                                                       dim=1)
                else:
                    # deep copy because we will modify the previous generated tokens to remove the signal tokens
                    final_generated_ids = generated_ids.clone()
                # remove the signal tokens
                generated_ids[cell_indices, signal_token_indices] = pad_token_id
                shift = torch.arange(num_signal_tokens).to(
                    device=generated_ids.device, dtype=generated_ids.dtype)
                indexer = signal_token_indices[:, None] + shift[None]
                final_generated_ids[
                    cell_indices[:, None].expand_as(indexer),
                    indexer] = shift[None] + first_signal_token_index

                # after the all signal tokens are added, we get the corresponding hidden states
                if self.base_model.config.is_encoder_decoder:
                    decoder_input_ids = final_generated_ids
                    last_hidden_states = self.base_model(
                        inputs_embeds=inputs_embeds,
                        decoder_input_ids=decoder_input_ids,
                        return_dict=True,
                        output_hidden_states=True,
                    ).decoder_hidden_states[-1]
                    condition_embeds = self._get_condition_embeds(
                        last_hidden_states, decoder_input_ids)
                elif self.base_model.config.is_decoder:
                    # unlike encoder-decoder architecture, the original input_ids are concatenated with the generated_ids
                    # note that we should consider padding tokens in the original input_ids
                    input_length = input_ids.size(-1)
                    new_input_ids = input_ids.clone()
                    num_normal_tokens = torch.sum(input_ids[cell_indices]
                                                  != pad_token_id,
                                                  dim=-1).to(dtype=torch.long)
                    max_num_normal_tokens = num_normal_tokens.max()
                    max_num_generated_tokens = (signal_token_indices.max() +
                                                num_signal_tokens)
                    # note that real nb_pad may be less than the calculated nb_pad
                    # because for an input with maximum length, its generated tokens may not be long enough to reach the maximum length
                    # in this case, we do it for the convenience of parallel computing
                    # we will remove the redundant padding tokens later
                    nb_pad = max_num_normal_tokens.item(
                    ) + max_num_generated_tokens.item() - input_length
                    if nb_pad > 0:
                        new_input_ids = torch.concat([
                            new_input_ids,
                            torch.full((batch_size, nb_pad), pad_token_id).to(
                                device=new_input_ids.device,
                                dtype=new_input_ids.dtype)
                        ],
                                                     dim=1)
                    shift = torch.arange(0, max_num_generated_tokens.item()).to(
                        device=num_normal_tokens.device, dtype=torch.long)
                    src_col_indexer = num_normal_tokens[:, None] + shift[None]
                    # concatenate the original input_ids with the generated_ids including the signal tokens
                    new_input_ids[
                        cell_indices[:, None].expand_as(src_col_indexer),
                        src_col_indexer] = final_generated_ids[cell_indices][
                            0:max_num_generated_tokens.item()]
                    # remove the redundant padding tokens
                    new_input_ids = new_input_ids[:,
                                                  torch.any(new_input_ids !=
                                                            pad_token_id,
                                                            dim=0)]
                    new_inputs_embeds = self.base_model.get_input_embeddings()(
                        new_input_ids)
                    last_hidden_states = self.base_model(
                        inputs_embeds=new_inputs_embeds,
                        return_dict=True,
                        output_hidden_states=True,
                    ).hidden_states[-1]
                    condition_embeds = self._get_condition_embeds(
                        last_hidden_states, new_input_ids)
                else:
                    raise ValueError(
                        "The model must be either encoder-decoder or decoder-only architecture."
                    )

                # use downstream model to generate the cell
                cells = self.feature_decoder.generate(condition_embeds,
                                                      to_numpy=True)
                for cell_indice, cell in zip(cell_indices, cells):
                    outputs["cells"][cell_indice] = cell
                if return_conditions:
                    condition_embeds_ = condition_embeds
                    if condition_embeds_.device.type == "cuda":
                        condition_embeds_ = condition_embeds_.cpu()
                    condition_embeds_ = condition_embeds_.numpy()
                    for cell_indice, condition in zip(cell_indices,
                                                      condition_embeds_):
                        outputs["conditions"][cell_indice] = condition

        if generated_ids.device.type == "cuda":
            generated_ids = generated_ids.cpu()
        outputs["texts"] = self.tokenizer.batch_decode(generated_ids,
                                                       skip_special_tokens=True)

        return outputs

    def forward(self,
                input_ids: torch.Tensor,
                attention_mask: Optional[torch.Tensor] = None,
                input_counts: Optional[torch.Tensor] = None,
                output_counts: Optional[torch.Tensor] = None,
                labels: Optional[torch.Tensor] = None,
                output_attentions: Optional[bool] = None,
                output_hidden_states: Optional[bool] = None,
                generative_model_kwargs: Optional[Dict[str, Any]] = None,
                **base_model_kwargs) -> Seq2SeqLMOutput | CausalLMOutput:
        """
        Perform a forward pass for the CellTextLLM model, integrating cell feature generation
        and text generation in a multimodal setup.

        Parameters
        ----------
        input_ids: torch.Tensor
            The token IDs of the input text sequence of shape (batch_size, sequence_length).
        attention_mask: torch.Tensor, optional, default None
            A mask that indicates which tokens should be attended to, where 1 indicates a token should be attended
            to and 0 indicates it should not be. If None, it is computed from ``input_ids``.
        input_counts: torch.Tensor, optional, default None
            A tensor of shape (num_cells, count_dim) representing the input cell counts to be processed by the feature encoder.
        output_counts: torch.Tensor, optional, default None
            A tensor representing expected output cell counts for supervision in the cell generation process.
        labels: torch.Tensor, optional, default None
            Ground truth labels for text generation. Used for supervised training when provided.
        output_attentions: bool, optional, default None
            If set to True, the model returns attention weights in addition to the output.
        output_hidden_states: bool, optional, default None
            If set to True, the model returns hidden states in addition to the output.
        generative_model_kwargs: dict, optional, default None
            Additional keyword arguments passed to the ``forward`` method of cell generative model if output cell counts
            are provided.
        **base_model_kwargs: dict
            Additional keyword arguments passed to the ``forward`` method of base language model.

        Returns
        -------
        outputs: transformers.modeling_outputs.Seq2SeqLMOutput or transformers.modeling_outputs.CausalLMOutput
            The outputs of the model.
        """
        # get text embeddings
        inputs_embeds = self.base_model.get_input_embeddings()(input_ids)
        if input_counts is not None:
            if self.normalize_total:
                input_counts = input_counts / input_counts.sum(
                    dim=-1, keepdim=True) * 1e4
            if self.log_variational:
                input_counts = torch.log1p(input_counts)
        if attention_mask is None:
            attention_mask = (input_ids
                              != self.base_model.config.pad_token_id).to(
                                  input_ids.device)
            attention_mask = attention_mask.long()

        # get cell modality's features
        if input_counts is not None:
            assert self.feature_encoder is not None, "Feature encoder is not provided."
            features = self.feature_encoder(input_counts)
            merged_inputs = self._merge_input_ids_with_features(
                input_ids,
                inputs_embeds,
                features,
                attention_mask,
                labels=labels,
            )
            inputs_embeds = merged_inputs["inputs_embeds"]
            attention_mask = merged_inputs["attention_mask"]
            if merged_inputs["labels"] is not None:
                labels = merged_inputs["labels"]

        base_model_kwargs = {
            **base_model_kwargs,
            "return_dict": True,
        }
        if output_counts is None:
            outputs = self.base_model(
                attention_mask=attention_mask,
                inputs_embeds=inputs_embeds,
                labels=labels,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
                **base_model_kwargs,
            )
            return outputs
        base_model_kwargs = {
            **base_model_kwargs,
            "output_hidden_states": True,
        }
        outputs = self.base_model(
            attention_mask=attention_mask,
            inputs_embeds=inputs_embeds,
            labels=labels,
            output_attentions=output_attentions,
            **base_model_kwargs,
        )

        # get conditional embeddings
        if hasattr(outputs, "decoder_hidden_states"):
            if "decoder_input_ids" not in base_model_kwargs:
                raise ValueError(
                    "decoder_input_ids must be provided in base_model_kwargs when the encoder-decoder model "
                    " is used to generate cells.")
            last_hidden_states = outputs.decoder_hidden_states[-1]
            input_ids_ = base_model_kwargs["decoder_input_ids"]
        else:
            last_hidden_states = outputs.hidden_states[-1]
            input_ids_ = input_ids
        condition_embeds = self._get_condition_embeds(
            last_hidden_states,
            input_ids_,
            output_counts=output_counts,
        )
        if generative_model_kwargs is None:
            generative_model_kwargs = {}
        generative_outputs = self.feature_decoder(condition_embeds,
                                                  output_counts,
                                                  **generative_model_kwargs)
        batch_size = input_ids.shape[0]
        num_generative_samples = output_counts.shape[0]
        outputs.loss = outputs.loss + generative_outputs.loss * num_generative_samples / batch_size

        return outputs


class InstructCellConfig(PretrainedConfig):
    """
    A configuration class for the ``InstructCell`` model, which integrates a base language model
    configuration, a feature encoder configuration, and a feature decoder configuration for
    multimodal cell-language tasks.

    Parameters
    ----------
    base_model_config: transformers.PretrainedConfig, optional, default None
        The configuration of the base language model. If this argument is a dictionary, the
        corresponding configuration class will be instantiated based on the ``model_type`` key
        in the dictionary. If it is None, a default configuration is created (e.g., T5).
    feature_encoder_config: Dict[str, Any], default {
        "is_q_former_encoder": True,
        "count_dim": 18961,
        "cross_attention_frequency": 1,
        "num_hidden_layers": 4,
        "num_key_value_tokens": 6,
        "num_blocks": 3,
        "num_query_tokens": 8,
        "hidden_dropout_prob": 0.1,
    }
        The configuration dictionary for the feature encoder. This dictionary can specify
        whether a Q-Former module is used, the dimensionality of the input cell counts,
        dropout probabilities, and other hyperparameters for the encoder architecture.
    feature_decoder_config: Dict[str, Any], default {
        "use_layer_norm": "both",
        "use_batch_norm": "none",
        "n_latent": 256,
        "condition_input_dim": 256,
        "log_variational": True,
        "n_layers": 4,
        "n_hidden": 1024,
        "dropout_rate": 0.1,
        "adaptive_library": True,
    }
        The configuration dictionary for the feature decoder. This dictionary can specify
        normalization schemes, number of latent dimensions, dimensionality of condition
        inputs, number of hidden layers, dropout rate, and whether library size is adaptive.
    tokenizer_use_fast: bool, default False
        Whether to use the fast tokenizer implementation from ``transformers`` library.
    normalize_total: bool, default True
        Whether to normalize total read count for each single-cell input.
    log_variational: bool, default True
        Whether to apply the log1p-transformation to to each single-cell input.
    **kwargs
        Additional keyword arguments passed to ``transformers.PretrainedConfig``.
    """

    model_type = "instructcell"
    sub_configs = {"base_model_config": AutoConfig}

    def __init__(
        self,
        base_model_config: Optional[PretrainedConfig] = None,
        feature_encoder_config: Dict[str, Any] = {
            "is_q_former_encoder": True,
            "count_dim": 18961,
            "cross_attention_frequency": 1,
            "num_hidden_layers": 4,
            "num_key_value_tokens": 6,
            "num_blocks": 3,
            "num_query_tokens": 8,
            "hidden_dropout_prob": 0.1,
        },
        feature_decoder_config: Dict[str, Any] = {
            "use_layer_norm": "both",
            "use_batch_norm": "none",
            "n_latent": 256,
            "condition_input_dim": 256,
            "log_variational": True,
            "n_layers": 4,
            "n_hidden": 1024,
            "dropout_rate": 0.1,
            "adaptive_library": True,
        },
        tokenizer_use_fast: bool = False,
        normalize_total: bool = True,
        log_variational: bool = True,
        **kwargs,
    ) -> "InstructCellConfig":
        # initialize the configuration of base language model
        if isinstance(base_model_config, dict):
            base_model_config["model_type"] = (base_model_config["model_type"]
                                               if "model_type"
                                               in base_model_config else "t5")
            base_model_config = CONFIG_MAPPING[base_model_config["model_type"]](
                **base_model_config)
        elif base_model_config is None:
            base_model_config = CONFIG_MAPPING["t5"]()
        self.base_model_config = base_model_config
        for attr_names in [
                "special_tokens_index_dict",
                "pad_token_id",
                "eos_token_id",
                "ignore_index",
                "num_signal_tokens",
        ]:
            if not hasattr(self.base_model_config, attr_names):
                self.base_model_config.__setattr__(attr_names, None)

        # initialize the configuration of feature encoder
        self.feature_encoder_config = feature_encoder_config

        # initialize the configuration of feature decoder
        self.feature_decoder_config = feature_decoder_config

        # the configuration of tokenizer
        self.tokenizer_use_fast = tokenizer_use_fast

        # other parameters of InstructCell
        self.normalize_total = normalize_total
        self.log_variational = log_variational

        super().__init__(**kwargs)


class InstructCell(PreTrainedModel):
    """
    A model class that integrates a base language model, a feature encoder, and a feature
    decoder to support multimodal interaction between text and single-cell data.

    The ``InstructCell`` model contains:
    - A base language model (e.g., T5, GPT).
    - A feature encoder (e.g., SCQFormer) to encode the input cell counts.
    - A feature decoder (e.g., Generator) to reconstruct or generate cell data (gene expression)
      from sampled latent vectors and given biological conditions.

    This class leverages a ``CellTextLLM`` instance, referred to as the assistant, to combine
    text embeddings and cell embeddings in a single forward pass or generation pipeline.
    It provides methods for tokenizing text, performing multimodal forward passes, and
    generating text or cell data outputs.

    Parameters
    ----------
    config: InstructCellConfig
        The configuration object containing hyperparameters and settings for the base model,
        feature encoder, feature decoder, tokenizer usage, and other relevant attributes.

    Notes
    -----
    - The model relies on a placeholder or signal tokens to identify where the cell embeddings
      should be inserted. Therefore, the user must invoke ``initialize_special_tokens``
      to set up these tokens in the tokenizer and in the model configuration if cell-based
      functionalities are needed.
    - The feature encoder and decoder used here typically expect preprocessed cell data,
      such as counts that may be normalized or log1p-transformed (controlled by
      ``config.normalize_total`` and ``config.log_variational``).
    """

    config_class = InstructCellConfig

    def __init__(self, config: InstructCellConfig) -> "InstructCell":
        # load the pretrained language model
        if config.base_model_config.is_encoder_decoder:
            model = AutoModelForSeq2SeqLM.from_config(config.base_model_config)
        else:
            model = AutoModelForCausalLM.from_config(config.base_model_config)

        # load the Q-former module
        count_dim = config.feature_encoder_config["count_dim"]
        is_q_former_encoder = config.feature_encoder_config[
            "is_q_former_encoder"]
        if is_q_former_encoder:
            cross_attention_frequency = config.feature_encoder_config[
                "cross_attention_frequency"]
            num_hidden_layers = config.feature_encoder_config[
                "num_hidden_layers"]
            qformer_config = Blip2QFormerConfig(
                vocab_size=0,
                hidden_size=model.config.hidden_size,
                hidden_dropout_prob=config.
                feature_encoder_config["hidden_dropout_prob"],
                num_hidden_layers=num_hidden_layers,
                num_attention_heads=model.config.num_attention_heads,
                intermediate_size=model.config.hidden_size * 4,
                pad_token_id=model.config.pad_token_id,
                cross_attention_frequency=cross_attention_frequency,
                encoder_hidden_size=model.config.hidden_size,
            )
            num_key_value_tokens = config.feature_encoder_config[
                "num_key_value_tokens"]
            num_blocks = config.feature_encoder_config["num_blocks"]
            num_query_tokens = config.feature_encoder_config["num_query_tokens"]
            feature_encoder = SCQFormer(
                count_dim,
                num_query_tokens,
                num_key_value_tokens,
                qformer_config,
                num_hidden_layers=num_blocks,
            )
        else:
            feature_encoder = nn.Sequential(
                nn.Linear(count_dim,
                          (count_dim + model.config.hidden_size) // 2),
                nn.GELU(),
                nn.Linear((count_dim + model.config.hidden_size) // 2,
                          model.config.hidden_size),
                nn.Dropout(
                    config.feature_encoder_config["hidden_dropout_prob"]),
            )

        # load the cell reconstruction module
        condition_input_dim = config.feature_decoder_config[
            "condition_input_dim"]
        use_layer_norm = config.feature_decoder_config["use_layer_norm"]
        use_batch_norm = config.feature_decoder_config["use_batch_norm"]
        n_latent = config.feature_decoder_config["n_latent"]
        log_variational = config.feature_decoder_config["log_variational"]
        n_layers = config.feature_decoder_config["n_layers"]
        n_hidden = config.feature_decoder_config["n_hidden"]
        dropout_rate = config.feature_decoder_config["dropout_rate"]
        adaptive_library = config.feature_decoder_config["adaptive_library"]
        # create a dummy matrix to initialize library size
        dummy_counts_for_init = np.random.randint(0, 1000, (10, count_dim))
        library_log_means, library_log_vars = init_library_size(
            dummy_counts_for_init)
        feature_decoder = Generator(
            count_dim,
            condition_dim=model.config.hidden_size,
            condition_input_dim=condition_input_dim,
            n_layers=n_layers,
            n_hidden=n_hidden,
            n_latent=n_latent,
            dropout_rate=dropout_rate,
            use_layer_norm=use_layer_norm,
            use_batch_norm=use_batch_norm,
            encode_covariates=True,
            deeply_inject_covariates=False,
            log_variational=log_variational,
            adaptive_library=adaptive_library,
            use_observed_lib_size=False,
            library_log_means=library_log_means,
            library_log_vars=library_log_vars,
        )

        super().__init__(config)
        # the submodule of InstructCell
        self.assistant = CellTextLLM(
            model,
            None,
            feature_encoder=feature_encoder,
            feature_decoder=feature_decoder,
            normalize_total=config.normalize_total,
            log_variational=config.log_variational,
        )

    def _check_tokenizer(self) -> None:
        """Check whether the tokenizer is set."""
        if self.assistant.tokenizer is None:
            raise ValueError(
                "The tokenizer has not been set. Please configure the tokenizer first."
            )

    def save_pretrained(self,
                        save_directory: Optional[str],
                        save_tokenizer_kwargs: Dict[str, Any] = {},
                        **kwargs) -> None:
        """
        Save the ``InstructCell`` model and its tokenizer to a directory, so that it can be
        reloaded using the ``from_pretrained`` method.

        Parameters
        ----------
        save_directory: str, optional
            The directory where the model and tokenizer files will be saved.
        save_tokenizer_kwargs: dict, default {}
            Additional keyword arguments passed to the tokenizer's ``save_pretrained`` method.
        **kwargs
            Additional keyword arguments passed to the ``save_pretrained`` method of the parent
            class (``transformers.PreTrainedModel``).
        """
        self._check_tokenizer()
        self.assistant.tokenizer.save_pretrained(save_directory,
                                                 **save_tokenizer_kwargs)
        super().save_pretrained(save_directory, **kwargs)

    @classmethod
    def from_pretrained(cls,
                        pretrained_model_name_or_path: Optional[str],
                        load_tokenizer_kwargs: Dict[str, Any] = {},
                        **kwargs) -> "InstructCell":
        """
        Load a pre-trained ``InstructCell`` model from a local directory or remote model hub.

        This method loads both the model weights and its tokenizer from the specified
        directory or model repository.

        Parameters
        ----------
        pretrained_model_name_or_path: str, optional
            The path (local directory) or name of the remote repository to load the model from.
        load_tokenizer_kwargs: Dict[str, Any], default {}
            Additional keyword arguments passed to the ``from_pretrained`` method of the class
            ``transformers.PreTrainedTokenizer`` or ``transformers.PreTrainedTokenizerFast``.
        **kwargs
            Additional keyword arguments passed to the ``from_pretrained`` method of the parent
            class (``transformers.PreTrainedModel``).

        Returns
        -------
        model: InstructCell
            The InstructCell model with trained base language model, feature encoder, feature
            decoder, and tokenizer.
        """
        model = super().from_pretrained(pretrained_model_name_or_path, **kwargs)
        load_tokenizer_kwargs = {
            "use_fast": model.config.tokenizer_use_fast,
            **load_tokenizer_kwargs
        }
        # load the corresponding tokenizer
        model.assistant.tokenizer = AutoTokenizer.from_pretrained(
            pretrained_model_name_or_path, **load_tokenizer_kwargs)
        return model

    def initialize_special_tokens(
            self,
            modality_tag: str = "CELL",
            num_signal_tokens: int = 1,
            pad_to_multiple_of: Optional[int] = 8) -> None:
        """
        Initialize and register special tokens for cell-language modeling, such as placeholder
        tokens and signal tokens for inserting or extracting cell embeddings.

        Parameters
        ----------
        modality_tag: str, default "CELL"
            The tag that represents the modality to be added to the tokenizer. It will be used to create
            start and end tags as well as signal tokens.
        num_signal_tokens: int, default 1
            The number of signal tokens to add. These tokens serve as signals enabling the model to generate
            data in the specified modality.
        pad_to_multiple_of: int, optional, default 8
            If set will pad the embedding matrix to a multiple of the provided value after resizing embedding
            layer. This is especially useful to enable the use of Tensor Cores on NVIDIA hardware with compute
            capability ``>= 7.5`` (Volta), or on TPUs which benefit from having sequence lengths be a multiple
            of 128.
        """
        self._check_tokenizer()
        # ensure the tokenizer has a pad token
        if self.assistant.tokenizer.pad_token_id is None:
            self.assistant.tokenizer.pad_token = self.assistant.tokenizer.eos_token
        ignore_index = -100 if not hasattr(self.assistant.base_model.config,
                                           "ignore_index") else None
        pad_token_id = self.assistant.tokenizer.pad_token_id if not hasattr(
            self.assistant.base_model.config, "pad_token_id") else None

        # prepare the model for cell-language modelling
        prepare_cell_text_llm(
            self.assistant.base_model,
            self.assistant.tokenizer,
            modality_tag=modality_tag,
            num_signal_tokens=num_signal_tokens,
            ignore_index=ignore_index,
            pad_token_id=pad_token_id,
            pad_to_multiple_of=pad_to_multiple_of,
        )

    def prepare_model_inputs(
        self,
        prompt: str,
        gene_counts: Optional[np.ndarray] = None,
        sc_metadata: Dict[str, str] = {},
    ) -> Dict[str, torch.Tensor]:
        """
        Prepare the input dictionary for the model, which includes tokenized text (``input_ids``)
        and optionally gene counts (``input_counts``).

        This method formats the prompt with single-cell metadata placeholders, tokenizes the
        resulting string, and wraps the input tokens in a PyTorch tensor. If gene counts are
        provided, they are also wrapped in a tensor.

        Parameters
        ----------
        prompt: str
            The text prompt that might include placeholders for single-cell metadata (e.g.,
            `{input}`, `{species}`). For example: ``"Please give the cell type of {input} from {species}."``
        gene_counts: np.ndarray, optional, default None
            The matrix of shape (num_cells, count_dim) or (count_dim,) that represents the
            input cell gene counts.
        sc_metadata: dict, default {}
            Additional metadata used to format the prompt string. For instance,
            ``{"species": "human"}``, which can be used to substitute for ``{species}`` in the prompt.
            By default, ``{input}`` in the prompt is automatically replaced with the special placeholder
            token (see method ``initialize_special_tokens``).

        Returns
        -------
        inputs: dict
            A dictionary containing:
            - "input_ids": torch.Tensor
              The tokenized text prompt of shape (1, sequence_length).
            - "input_counts": torch.Tensor or None
              The cell count tensor of shape (num_cells, count_dim) if gene_counts is not None,
              otherwise None.
        """
        if not hasattr(self.config.base_model_config, "special_tokens_dict"):
            raise ValueError(
                "Cannot get special tokens. ",
                "Please invoke `initialize_special_tokens` for cell-language modelling."
            )
        self._check_tokenizer()

        if gene_counts is not None and len(gene_counts.shape) == 1:
            gene_counts = gene_counts.reshape(1, -1)

        placeholder = self.config.base_model_config.special_tokens_dict[
            "placeholder"]
        start_tag = self.config.base_model_config.special_tokens_dict[
            "start_tag"]
        end_tag = self.config.base_model_config.special_tokens_dict["end_tag"]

        # users can define the value of key `input` in `sc_metadata` by themselves
        sc_metadata = {
            "input": f"{start_tag}{placeholder}{end_tag}",
            **sc_metadata,
        }
        prompt = prompt.format(**sc_metadata)
        prompt = f"User:\n{prompt}\n\nAssistant:\n"

        input_ids = self.assistant.tokenizer(prompt,
                                             return_tensors="np").input_ids
        inputs = {"input_ids": torch.from_numpy(input_ids)}
        if gene_counts is not None:
            inputs["input_counts"] = torch.from_numpy(gene_counts)
        else:
            inputs["input_counts"] = None
        for key, value in inputs.items():
            if value is not None:
                inputs[key] = value.to(device=self.assistant.base_model.device)
        if inputs["input_counts"] is not None:
            inputs["input_counts"] = inputs["input_counts"].to(dtype=self.dtype)

        return inputs

    def forward(self,
                input_ids: torch.Tensor,
                attention_mask: Optional[torch.Tensor] = None,
                input_counts: Optional[torch.Tensor] = None,
                output_counts: Optional[torch.Tensor] = None,
                labels: Optional[torch.Tensor] = None,
                output_attentions: Optional[bool] = None,
                output_hidden_states: Optional[bool] = None,
                generative_model_kwargs: Optional[Dict[str, Any]] = None,
                **base_model_kwargs) -> Seq2SeqLMOutput | CausalLMOutput:
        """
        Perform a forward pass for the CellTextLLM model, integrating cell feature generation
        and text generation in a multimodal setup.

        Parameters
        ----------
        input_ids: torch.Tensor
            The token IDs of the input text sequence of shape (batch_size, sequence_length).
        attention_mask: torch.Tensor, optional, default None
            A mask that indicates which tokens should be attended to, where 1 indicates a token should be attended
            to and 0 indicates it should not be. If None, it is computed from ``input_ids``.
        input_counts: torch.Tensor, optional, default None
            A tensor of shape (num_cells, count_dim) representing the input cell counts to be processed by the feature encoder.
        output_counts: torch.Tensor, optional, default None
            A tensor representing expected output cell counts for supervision in the cell generation process.
        labels: torch.Tensor, optional, default None
            Ground truth labels for text generation. Used for supervised training when provided.
        output_attentions: bool, optional, default None
            If set to True, the model returns attention weights in addition to the output.
        output_hidden_states: bool, optional, default None
            If set to True, the model returns hidden states in addition to the output.
        generative_model_kwargs: dict, optional, default None
            Additional keyword arguments passed to the ``forward`` method of cell generative model if output cell counts
            are provided.
        **base_model_kwargs: dict
            Additional keyword arguments passed to the ``forward`` method of base language model.

        Returns
        -------
        outputs: transformers.modeling_outputs.Seq2SeqLMOutput or transformers.modeling_outputs.CausalLMOutput
            The outputs of the model.
        """
        return self.assistant(
            input_ids,
            attention_mask=attention_mask,
            input_counts=input_counts,
            output_counts=output_counts,
            labels=labels,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            generative_model_kwargs=generative_model_kwargs,
            **base_model_kwargs,
        )

    def generate(
        self,
        input_ids: torch.Tensor,
        input_counts: Optional[torch.Tensor] = None,
        return_conditions: bool = False,
        attention_mask: Optional[torch.Tensor] = None,
        prefix_allowed_tokens_fn: Optional[Callable[[int, torch.Tensor],
                                                    List[int]]] = None,
        synced_gpus: Optional[bool] = None,
        assistant_model: Optional["PreTrainedModel"] = None,
        streamer: Optional["BaseStreamer"] = None,
        **kwargs,
    ) -> Dict[str, List[str | None | np.ndarray]]:
        """
        Generate texts or cells from text sequences ``input_ids`` and input counts ``input_counts`` (optionally).

        This method combines tokenized text input with encoded cell features to generate both text
        sequences and optionally cell data. It handles the integration of text and cell data using
        a feature encoder and decoder if they are provided.

        Parameters
        ----------
        input_ids: torch.Tensor
            Input token IDs of shape (batch_size, sequence_length) for the language model.
        input_counts: torch.Tensor, optional, default None
            Input counts of shape (num_cells, count_dim), which are passed through the feature encoder.
            If not provided, text-only generation is performed.
        return_conditions: bool, optional, default False
            Whether to return the condition embeddings (i.e. the last hidden states of signal tokens).
        attention_mask: torch.Tensor, optional, default None
            A mask that specifies which tokens should be attended to. If not provided, it is computed from ``input_ids``.
        prefix_allowed_tokens_fn: Callable[[int, torch.Tensor], List[int]], optional, default None
            If provided, this function constraints the beam search to allowed tokens only at each step. If not provided,
            no constraint is applied. Please see
            https://huggingface.co/docs/transformers/v4.44.2/en/main_classes/text_generation#transformers.GenerationMixin.generate
            for more details.
        synced_gpus: bool, optional, default None
            Whether to continue running the while loop until max_length. Please see
            https://huggingface.co/docs/transformers/v4.44.2/en/main_classes/text_generation#transformers.GenerationMixin.generate
            for more details.
        assistant_model: PreTrainedModel, optional, default None
            An assistant model that can be used to accelerate generation. The assistant model must have the exact same tokenizer.
            The acceleration is achieved when forecasting candidate tokens with the assistent model is much faster than running
            generation with the model you are calling generate from. As such, the assistant model should be much smaller.
            For more details, please see https://arxiv.org/abs/2302.01318.
        streamer: BaseStreamer, optional, default None
            Streamer object that will be used to stream the generated sequences. Generated tokens are passed through
            ``streamer.put(token_ids)`` and the streamer is responsible for any further processing.
        **kwargs
            Additional keyword arguments passed to the ``generate`` method of the language model. In general,
            the language model is based on the transformers library. Thus, please see
            https://huggingface.co/docs/transformers/v4.44.2/en/main_classes/text_generation#transformers.GenerationMixin.generate
            for more details.

        Returns
        -------
        outputs: dict
            A dict containing the following keys:
            - "texts": list of str, the generated text sequences. Special tokens are removed.
            - "cells": list of cell data, where each entry is the decoded cell features. If no signal tokens are detected during
            generation, each entry is None.
            - "conditions": list of numpy array, where each entry is the condition embeddings. If no signal tokens are detected during
            generation, each entry is None. This key is only present if ``return_conditions`` is True.
        """
        self._check_tokenizer()
        return self.assistant.generate(
            input_ids,
            input_counts=input_counts,
            return_conditions=return_conditions,
            attention_mask=attention_mask,
            prefix_allowed_tokens_fn=prefix_allowed_tokens_fn,
            synced_gpus=synced_gpus,
            assistant_model=assistant_model,
            streamer=streamer,
            **kwargs,
        )

    def predict(self,
                prompt: str,
                gene_counts: Optional[np.ndarray] = None,
                sc_metadata: Dict[str, str] = {},
                **kwargs) -> Dict[str, str | np.ndarray]:
        """
        A high-level helper method for inference, combining text prompt processing and generation
        in a single call.

        This method:
        1. Prepares the model inputs using ``prepare_model_inputs`` (tokenizes the prompt and
           optionally includes the cell counts).
        2. Calls the model's ``generate`` method to produce text (and optionally cell data).
        3. Returns the generated text and cell data in a dictionary.

        Parameters
        ----------
        prompt: str
            The user prompt or instruction, which may contain placeholders for single-cell data.
        gene_counts: np.ndarray, optional, default None
            The cell count array of shape (num_cells, count_dim) or (count_dim,).
        sc_metadata: Dict[str, str], default {}
            Additional metadata to format the prompt. Keys can be placeholders used in ``prompt``.
        **kwargs
            Additional keyword arguments passed to the model's ``generate`` method
            (e.g., ``top_p``).

        Returns
        -------
        outputs: dict
            A dictionary containing:
            - "text": str, the generated text in response to the prompt.
            - "cell": np.ndarray or None, the generated cell data, if applicable.

        Notes
        -----
        - This method is particularly convenient for quick inference scenarios where the user
          has cell data (counts) and a text prompt, and wants a single call to retrieve both
          text and cell outputs.
        """
        inputs = self.prepare_model_inputs(prompt,
                                           gene_counts=gene_counts,
                                           sc_metadata=sc_metadata)
        outputs = self.generate(inputs["input_ids"],
                                input_counts=inputs["input_counts"],
                                **kwargs)
        return {"text": outputs["texts"][0], "cell": outputs["cells"][0]}



class scVILoader():

    def __init__(self):
        pass

    def load(self, census_version):
        import requests
        import os

        scvi_url = f"https://cellxgene-contrib-public.s3.us-west-2.amazonaws.com/models/scvi/{census_version}/homo_sapiens/model.pt"
        os.makedirs(os.path.join(os.getcwd(), 'scvi_model'), exist_ok=True)

        output_path = os.path.join('scvi_model', 'model.pt')

        try:
            response = requests.get(scvi_url, verify=False)
            if response.status_code == 404:
                raise Exception(
                    'Census version not found, defaulting to version 2024-07-01'
                )
        except Exception as e:
            print(e)
            census_version = "2024-07-01"
            scvi_url = f"https://cellxgene-contrib-public.s3.us-west-2.amazonaws.com/models/scvi/2024-07-01/homo_sapiens/model.pt"
            response = requests.get(scvi_url, verify=False)

        with open(output_path, "wb") as file:
            file.write(response.content)

        print(
            f'scVI version {census_version} downloaded to {output_path} in current directory'
        )



# -*- coding: utf-8 -*-

from __future__ import division
from __future__ import print_function

import os
import sys

import unittest
import shutil

# temporary solution for relative imports in case TDC is not installed
# if TDC is installed, no need to use the following line
sys.path.append(
    os.path.abspath(os.path.join(os.path.dirname(__file__), "../..")))


class TestFunctions(unittest.TestCase):

    def setUp(self):
        print(os.getcwd())
        pass

    def test_Evaluator(self):
        from tdc_ml import Evaluator

        evaluator = Evaluator(name="ROC-AUC")
        print(evaluator([0, 1], [0.5, 0.6]))

    def test_binarize(self):
        from tdc_ml.single_pred import TestSinglePred

        data = TestSinglePred(name="Test_Single_Pred")
        data.binarize(threshold=-5, order="descending")

    def test_convert_to_log(self):
        from tdc_ml.single_pred import TestSinglePred

        data = TestSinglePred(name="Test_Single_Pred")
        data.convert_to_log()

    def test_print_stats(self):
        from tdc_ml.single_pred import TestSinglePred

        data = TestSinglePred(name="Test_Single_Pred")
        data.print_stats()

    def tearDown(self):
        print(os.getcwd())

        if os.path.exists(os.path.join(os.getcwd(), "data")):
            shutil.rmtree(os.path.join(os.getcwd(), "data"))
        if os.path.exists(os.path.join(os.getcwd(), "oracle")):
            shutil.rmtree(os.path.join(os.getcwd(), "oracle"))


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

from __future__ import division
from __future__ import print_function

import os
import sys

import unittest
import shutil

# temporary solution for relative imports in case TDC is not installed
# if TDC is installed, no need to use the following line
sys.path.append(
    os.path.abspath(os.path.join(os.path.dirname(__file__), "../..")))
# TODO: add verification for the generation other than simple integration


class TestDataloader(unittest.TestCase):

    def setUp(self):
        print(os.getcwd())
        pass

    def test_single_pred(self):
        from tdc_ml.single_pred import TestSinglePred

        data = TestSinglePred(name="Test_Single_Pred")
        split = data.get_split()

    def test_multi_pred(self):
        from tdc_ml.multi_pred import TestMultiPred

        data = TestMultiPred(name="Test_Multi_Pred")
        split = data.get_split()

    def test_brown(self):
        # TODO: factor out into specialized test suites for individual datasets
        # this test serves as an integration test of the data processing, data configs, and existing tdc pipeline. leave here for now.
        from tdc_ml.multi_pred import ProteinPeptide
        data = ProteinPeptide(name="brown_mdm2_ace2_12ca5")
        assert "protein_or_rna_sequence" in data.get_data(
        ).columns  # brown protein<>peptide dataset uses a data config inserting this column
        data.get_split()

    @unittest.skip(
        "test is taking up too much memory"
    )  #FIXME: should probably create much smaller version and use that for the test. This test does pass locally, please rerun if changing anndata code.
    def test_h5ad_dataloader(self):
        from tdc_ml.multi_pred.perturboutcome import PerturbOutcome
        from pandas import DataFrame
        test_loader = PerturbOutcome(
            name="scperturb_drug_AissaBenevolenskaya2021")
        testdf = test_loader.get_data()
        assert isinstance(testdf, DataFrame)
        test_loader.get_split()

    def test_generation(self):
        from tdc_ml.generation import MolGen

        data = MolGen(name="ZINC")
        split = data.get_split()

    def test_resource_dataverse_dataloader(self):
        import pandas as pd
        from tdc_ml.resource.dataloader import DataLoader

        data = DataLoader(name="opentargets_dti")
        df = data.get_data()
        assert "Y" in df.columns
        split = data.get_split()
        assert "train" in split
        assert len(split["train"]) > 0
        assert len(split["test"]) > 0
        assert isinstance(split["train"], pd.DataFrame)

    def test_resource_dataverse_dataloader_raw_splits(self):
        import pandas as pd
        from tdc_ml.resource.dataloader import DataLoader
        data = DataLoader(name="tchard")
        df = data.get_data()
        assert isinstance(df, pd.DataFrame)
        assert "Y" in df.columns
        assert "splits" in data
        splits = data.get_split()
        assert "train" in splits
        assert "dev" in splits
        assert "test" in splits
        assert isinstance(
            splits["train"]["tchard_pep_cdr3b_only_neg_assays"][0],
            pd.DataFrame)
        assert isinstance(splits["test"]["tchard_pep_cdr3b_only_neg_assays"][2],
                          pd.DataFrame)
        assert not splits["dev"]

    def test_mpc(self):
        from tdc_ml.single_pred.mpc import MPC
        Xs = MPC(
            name=
            "https://raw.githubusercontent.com/bidd-group/MPCD/main/dataset/ADMET/DeepDelta_benchmark/Caco2.csv"
        )
        Xs_split = Xs.get_split()
        Xs_train = Xs_split["train"]
        Xs_test = Xs_split["test"]
        y_train_pIC50 = Xs_train["Y"]
        y_test_pIC50 = Xs_test["Y"]

    def tearDown(self):
        try:
            print(os.getcwd())
            shutil.rmtree(os.path.join(os.getcwd(), "data"))
        except:
            pass


if __name__ == "__main__":
    unittest.main()



import unittest
import shutil
import numpy as np
import random
import sys
import os

sys.path.append(
    os.path.abspath(os.path.join(os.path.dirname(__file__), "../..")))
from tdc_ml.benchmark_group import admet_group, scdti_group, counterfactual_group, geneperturb_group


def is_classification(values):
    value_set = set(values)
    if len(value_set) == 2:
        return True
    return False


class TestBenchmarkGroup(unittest.TestCase):

    def setUp(self):
        self.group = admet_group(path="data/")

    def tearDown(self) -> None:
        shutil.rmtree("data", ignore_errors=True)

    def test_ADME_mean_prediction(self):
        predictions = {}
        num_datasets = 0
        for my_group in self.group:
            num_datasets += 1
            name = my_group["name"]
            test = my_group["test"]
            mean_val = np.mean(test["Y"])
            predictions[name] = [mean_val] * len(test)

        results = self.group.evaluate(predictions)
        self.assertEqual(num_datasets, len(results))
        for my_group in self.group:
            self.assertTrue(my_group["name"] in results)

    def test_ADME_evaluate_many(self):
        prediction_list = []
        for random_seed in range(5):
            predictions = {}
            for my_group in self.group:
                name = my_group["name"]
                test = my_group["test"]
                predictions[name] = test["Y"]
            prediction_list.append(predictions)

        results = self.group.evaluate_many(prediction_list)
        for ds_name, metrics in results.items():
            self.assertEqual(len(metrics), 2)
            u, std = metrics
            self.assertTrue(u
                            in (1,
                                0))  # A perfect score for all metrics is 1 or 0
            self.assertEqual(0, std)

        for my_group in self.group:
            self.assertTrue(my_group["name"] in results)

    def test_SCDTI_benchmark(self):
        group = scdti_group.SCDTIGroup()
        train_val = group.get_train_valid_split()
        assert "train" in train_val, "no training set"
        assert "val" in train_val, "no validation set"
        assert len(train_val["train"]) > 0, "no entries in training set"
        tst = group.get_test()["test"]
        tst["preds"] = tst["y"]  # switch predictions to ground truth
        results = group.evaluate(tst)
        assert "IBD" in results, "missing ibd from diseases. got {}".format(
            results.keys())
        assert "RA" in results, "missing ra from diseases. got {}".format(
            results.keys())
        assert results["IBD"] == results[
            "RA"], "both should be perfect scores but got IBD {} vs RA {}".format(
                results["IBD"], results["RA"])  # both should be perfect scores
        assert results["IBD"] - 1.0 < 0.000001  # should be a perfect score
        many_results = group.evaluate_many([tst] * 5)
        assert "IBD" in many_results, "missing ibd from diseases in evaluate many. got {}".format(
            many_results.keys())
        assert "RA" in many_results, "missing ra from diseases in evaluate many. got {}".format(
            many_results.keys())
        assert len(many_results["IBD"]) == len(
            many_results["RA"]
        ), "both diseases should include mean and standard deviation"
        assert len(many_results["IBD"]
                  ) == 2, "results should include mean and standard deviation"
        assert many_results["IBD"][
            0] - 1.0 < 0.000001, "should get perfect score"

    @unittest.skip(
        "counterfactual test is taking up too much memory"
    )  #FIXME: please run if making changes to counterfactual benchmark or core code.
    def test_counterfactual(self):
        from tdc_ml.multi_pred.perturboutcome import PerturbOutcome
        from tdc_ml.dataset_configs.config_map import scperturb_datasets, scperturb_gene_datasets

        test_data = PerturbOutcome("scperturb_drug_AissaBenevolenskaya2021")
        print("got test data")
        group = counterfactual_group.CounterfactualGroup()  # is drug
        assert group.is_drug
        assert set(group.dataset_names) == set(
            scperturb_datasets
        ), "loaded datasets should be scperturb drug, but were {} vs correct: {}".format(
            group.dataset_names, scperturb_datasets)
        print("getting test data via dataloader")
        ct = len(test_data.get_data())
        print("getting splits")
        train, val = group.get_train_valid_split(remove_unseen=False)
        test = group.get_test()
        print("got splits; checking counts")
        trainct = sum(len(x) for _, x in train.items())
        valct = sum(len(x) for _, x in val.items())
        testct = sum(len(x) for _, x in test.items())
        controlct = sum(len(x) for _, x in group.split["control"].items())
        adjct = group.split["adj"]
        totalct = trainct + valct + testct + controlct + adjct
        assert ct == totalct, "counts between original data and the 3 splits should match: original {} vs splits {}".format(
            ct, totalct)
        # basic test on perfect score
        print("benchmark - generating identical test set")
        tst = test_data.get_split(remove_unseen=False)
        tstdict = {}
        for line, splits in tst.items():
            tstdict[line] = splits["test"]
        print("benchmark - running evaluate")
        r2 = group.evaluate(tstdict)
        assert r2 == 1, "comparing test to itself should have perfect R^2 score, was {}".format(
            r2)
        print("done")
        # now just check we can load sc perturb gene correctly
        print("benchmark - basic load test on the gene benchmark")
        group_gene = counterfactual_group.CounterfactualGroup(is_drug=False)
        assert not group_gene.is_drug
        assert set(group_gene.dataset_names) == set(scperturb_gene_datasets)

    @unittest.skip(
        "counterfactual test is taking up too much memory"
    )  #FIXME: please run if making changes to counterfactual benchmark or core code.
    def test_gene_perturb(self):
        group = geneperturb_group.GenePerturbGroup()
        group.get_train_valid_split()
        group.get_test()

    @unittest.skip(
        "mygene dependency removal")  #FIXME: separate into conda-only tests
    def test_proteinpeptide(self):
        from tdc_ml.benchmark_group.protein_peptide_group import ProteinPeptideGroup
        from tdc_ml.multi_pred.proteinpeptide import ProteinPeptide
        from sklearn.model_selection import train_test_split
        group = ProteinPeptideGroup()
        test = group.get_test()
        assert test is not None and len(test) > 0
        dl = ProteinPeptide(name="brown_mdm2_ace2_12ca5")
        df = dl.get_data()
        for idx, e in enumerate(df["Y"]):
            if e != "Putative binder":
                df["Y"][idx] = "1"
            else:
                df["Y"][idx] = "0"
        # raise Exception("unique", )
        # Split the data while stratifying
        _, _, _, y_test = train_test_split(
            df.drop('Y', axis=1),  # features
            df['Y'],  # labels
            test_size=0.9,  # 90% of the data goes to the test set
            random_state=42,  # for reproducibility
            stratify=df[
                'Y']  # stratify by the label column to ensure even distribution
        )
        res = group.evaluate(y_test)
        assert res[-1] == 1 and res[-2] == 1, res

    def test_tcrepitope(self):
        from tdc_ml.benchmark_group.tcrepitope_group import TCREpitopeGroup
        from tdc_ml.resource.dataloader import DataLoader
        data = DataLoader("tchard")
        tst = data.get_split()["test"]
        group = TCREpitopeGroup()
        res = group.evaluate(tst)
        assert res == 1


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

from __future__ import division
from __future__ import print_function

import os
import sys

import unittest
import shutil

import pandas as pd

sys.path.append(
    os.path.abspath(os.path.join(os.path.dirname(__file__), "../..")))

from tdc_ml.feature_generators.data_feature_generator import DataFeatureGenerator
from tdc_ml.feature_generators.protein_feature_generator import ProteinFeatureGenerator


class TestDataFeatureGenerator(unittest.TestCase):

    def setUp(self):
        print(os.getcwd())
        pass

    def testAutofill(self):
        test_entries = [[0, "x", 8], [1, 'y', 4], [None, "x", 9],
                        [None, "y", 8], [2, "z", 12]]
        col_names = ["autofill", "index", "value"]
        df = pd.DataFrame(test_entries, columns=col_names)
        df2 = DataFeatureGenerator.autofill_identifier(df, "autofill", "index")
        self.assertEqual(df["autofill"].tolist(), [0, 1, 0, 1, 2])
        self.assertEqual(df2["autofill"].tolist(), [0, 1, 0, 1, 2])
        self.assertEqual(df2["index"].tolist(), ["x", "y", "x", "y", "z"])
        self.assertEqual(df2["value"].tolist(), [8, 4, 9, 8, 12])
        self.assertEqual(df2.shape[0], 5)
        self.assertEqual(df2.shape[1], 3)

    def testCreateRange(self):
        test_entries = [["7.7±4.5", 0], ["10±2.3", 1], ["Putative binder", 5]]
        col_names = ["num", "some_value"]
        keys = ["Putative binder"]
        subs = [0]
        df = pd.DataFrame(test_entries, columns=col_names)
        df2 = DataFeatureGenerator.create_range(df, "num", keys, subs)
        assert "expected" in df.columns
        assert "expected" in df2.columns
        assert "lower" in df2.columns
        assert "upper" in df2.columns
        self.assertEqual(df2["expected"].tolist(), [7.7, 10, 0])
        self.assertEqual(df2["lower"].tolist(), [3.2, 7.7, 0])
        self.assertEqual(df2["upper"].tolist(), [12.2, 12.3, 0])
        self.assertEqual(df2["num"].tolist(),
                         ["7.7±4.5", "10±2.3", "Putative binder"])
        self.assertEqual(df2["some_value"].tolist(), [0, 1, 5])
        self.assertEqual(df2.shape[0], 3)
        self.assertEqual(df2.shape[1], 5)

    def testProcessData(self):
        test_entries1 = [
            ["7.7±4.5", 0],
            ["10±2.3", 1],
            ["Putative binder", 5],
            ["Putative binder", 5],
            ["Putative binder", 5],
        ]
        test_entries2 = [[0, "x", 8], [1, 'y', 4], [None, "x", 9],
                         [None, "y", 8], [2, "z", 12]]
        test_entries = [x + y for x, y in zip(test_entries1, test_entries2)]
        col_names = [
            "num",
            "some_value",
            "autofill",
            "index",
            "value",
        ]
        assert len(col_names) == len(test_entries[0]),\
            ("Number of columns in test_entries does not match number of columns in col_names", col_names, test_entries[0])
        functions = [
            "create_range",
            "autofill_identifier",
        ]
        args = [
            {
                "column": "num",
                "keys": ["Putative binder"],
                "subs": [0],
            },
            {
                "autofill_column": "autofill",
                "key_column": "index"
            },
        ]
        df = pd.DataFrame(test_entries, columns=col_names)
        df2 = DataFeatureGenerator.process_data(df, functions, args)
        assert "expected" in df.columns
        assert "expected" in df2.columns
        assert "lower" in df2.columns
        assert "upper" in df2.columns
        self.assertEqual(df2["expected"].tolist(), [7.7, 10, 0, 0, 0])
        self.assertEqual(df2["lower"].tolist(), [3.2, 7.7, 0, 0, 0])
        self.assertEqual(df2["upper"].tolist(), [12.2, 12.3, 0, 0, 0])
        self.assertEqual(df2["num"].tolist(), [
            "7.7±4.5", "10±2.3", "Putative binder", "Putative binder",
            "Putative binder"
        ])
        self.assertEqual(df2["some_value"].tolist(), [0, 1, 5, 5, 5])
        self.assertEqual(df2.shape[0], 5)
        self.assertEqual(df2.shape[1], 8)
        self.assertEqual(df["autofill"].tolist(), [0, 1, 0, 1, 2])
        self.assertEqual(df2["autofill"].tolist(), [0, 1, 0, 1, 2])
        self.assertEqual(df2["index"].tolist(), ["x", "y", "x", "y", "z"])
        self.assertEqual(df2["value"].tolist(), [8, 4, 9, 8, 12])

    def tearDown(self):
        print(os.getcwd())

        if os.path.exists(os.path.join(os.getcwd(), "data")):
            shutil.rmtree(os.path.join(os.getcwd(), "data"))
        if os.path.exists(os.path.join(os.getcwd(), "oracle")):
            shutil.rmtree(os.path.join(os.getcwd(), "oracle"))


class TestProteinDataUtil(unittest.TestCase):

    def setUp(self):
        print(os.getcwd())
        pass

    def testInsertProteinSequence(self):
        test_entries = [
            ["BRCA1"],
            ["MDM2"],
            ["ACE2"],
            ["12CA5"],
        ]
        col_names = ["Gene name"]
        df = pd.DataFrame(test_entries, columns=col_names)
        df2 = ProteinFeatureGenerator.insert_protein_sequence(df, "Gene name")
        self.assertEqual(df2.shape[0], 4)
        self.assertEqual(df2.shape[1], 3)
        self.assertEqual(df2["Gene name"].tolist(),
                         ["BRCA1", "MDM2", "ACE2", "12CA5"])
        self.assertEqual(len(df2["protein_or_rna_sequence"]), 4)
        self.assertEqual(df2["protein_or_rna_sequence"].tolist(),
                         df["protein_or_rna_sequence"].tolist())
        self.assertEqual(len(df2["protein_or_rna_sequence"].unique()), 4)

    def tearDown(self):
        print(os.getcwd())

        if os.path.exists(os.path.join(os.getcwd(), "data")):
            shutil.rmtree(os.path.join(os.getcwd(), "data"))
        if os.path.exists(os.path.join(os.getcwd(), "oracle")):
            shutil.rmtree(os.path.join(os.getcwd(), "oracle"))


if __name__ == "__main__":
    unittest.main()



import os
import sys

sys.path.append(
    os.path.abspath(os.path.join(os.path.dirname(__file__), "../..")))
import unittest
import shutil

from pandas import DataFrame


class TestResources(unittest.TestCase):
    pass


class TestPrimeKG(unittest.TestCase):

    def setUp(self):
        print(os.getcwd())
        pass

    def test_node_retrieval(self):
        from tdc_ml.resource import PrimeKG
        data = PrimeKG(path='./data')
        drug_feature = data.get_features(feature_type='drug')
        data.to_nx()
        l = data.get_node_list('disease')
        assert "1" in l and "9997" in l


class TestPINNACLE(unittest.TestCase):

    def setUp(self):
        print(os.getcwd())
        pass

    def test_mg_ppi_load(self):
        from tdc_ml.resource.pinnacle import PINNACLE
        pinnacle = PINNACLE()
        assert isinstance(pinnacle.get_ppi(), DataFrame)
        assert isinstance(pinnacle.get_mg(), DataFrame)
        assert len(pinnacle.get_ppi()) > 0
        assert len(pinnacle.get_mg()) > 0
        embeds = pinnacle.get_embeds()
        assert isinstance(embeds, DataFrame)
        assert len(embeds) > 0, "PINNACLE embeds is empty"

    def test_embeddings(self):
        from tdc_ml.resource.pinnacle import PINNACLE
        pinnacle = PINNACLE()
        embeds = pinnacle.get_embeds()
        assert isinstance(embeds, DataFrame)
        assert len(embeds) > 0, "PINNACLE embeds is empty"
        keys = pinnacle.get_keys()
        assert isinstance(keys, DataFrame)
        assert len(keys) > 0, "PINNACLE keys is empty"
        assert len(keys) == len(embeds), "{} vs {}".format(
            len(keys), len(embeds))
        num_targets = len(keys["target"].unique())
        num_cells = len(keys["cell type"].unique())
        all_entries = embeds.index
        prots = [x.split("--")[0] for x in all_entries]
        cells = [x.split("--")[1] for x in all_entries]
        assert len(
            set(prots)) == num_targets, "{} vs {} for target proteins".format(
                len(prots), num_targets)
        assert len(set(cells)) == num_cells, "{} vs {} for cell_types".format(
            len(cells), num_cells)

    def test_exp_data(self):
        from tdc_ml.resource.pinnacle import PINNACLE
        pinnacle = PINNACLE()
        exp_data = pinnacle.get_exp_data()
        assert isinstance(exp_data, DataFrame)
        assert len(exp_data) > 0, "PINNACLE exp_data is empty"

    def tearDown(self):
        try:
            print(os.getcwd())
            shutil.rmtree(os.path.join(os.getcwd(), "data"))
        except:
            pass


class TestPharmoneMap(unittest.TestCase):

    def setUp(self):
        print(os.getcwd())
        pass

    def test_get_data(self):
        from tdc_ml.resource.pharmone import PharmoneMap

        resource = PharmoneMap()
        data = resource.get_data()
        assert isinstance(data, DataFrame), type(data)
        assert "Compound" in data.columns
        assert "Target_ID" in data.columns
        assert "pXC50" in data.columns

    def tearDown(self):
        try:
            print(os.getcwd())
            shutil.rmtree(os.path.join(os.getcwd(), "data"))
        except:
            pass


if __name__ == "__main__":
    unittest.main()



from .evaluator import (
    validity,
    uniqueness,
    novelty,
    diversity,
    kl_divergence,
    fcd_distance,
)
from .featurize.molconvert import MolConvert
from .oracle.filter import MolFilter



# -*- coding: utf-8 -*-

from __future__ import division
from __future__ import print_function

import os
import sys

import unittest
import shutil
import pytest

# temporary solution for relative imports in case TDC is not installed
# if TDC is installed, no need to use the following line
sys.path.append(
    os.path.abspath(os.path.join(os.path.dirname(__file__), "../..")))
# TODO: add verification for the generation other than simple integration


class TestHF(unittest.TestCase):

    def setUp(self):
        print(os.getcwd())
        pass

    @pytest.mark.skip(
        reason="This test is skipped due to deeppurpose installation dependency"
    )
    @unittest.skip(reason="DeepPurpose")
    def test_hf_load_predict(self):
        from tdc_ml.single_pred import Tox
        data = Tox(name='herg_karim')

        from tdc_ml import tdc_hf_interface
        tdc_hf = tdc_hf_interface("hERG_Karim-CNN")
        # load deeppurpose model from this repo
        dp_model = tdc_hf.load_deeppurpose('./data')
        tdc_hf.predict_deeppurpose(dp_model, ['CC(=O)NC1=CC=C(O)C=C1'])

    def test_hf_transformer(self):
        from tdc_ml import tdc_hf_interface
        # from transformers import Pipeline
        from transformers import BertForMaskedLM as BertModel
        geneformer = tdc_hf_interface("Geneformer")
        model = geneformer.load()
        # assert isinstance(pipeline, Pipeline)
        assert isinstance(model, BertModel), type(model)

    def tearDown(self):
        try:
            print(os.getcwd())
            shutil.rmtree(os.path.join(os.getcwd(), "data"))
        except:
            pass


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# Author: TDC Team
# License: MIT

import numpy as np
import os

try:
    from rdkit import Chem, DataStructs
    from rdkit.Chem import AllChem
    from rdkit import rdBase

    rdBase.DisableLog("rdApp.error")
except:
    raise ImportError(
        "Please install rdkit by 'conda install -c conda-forge rdkit'! ")


def single_molecule_validity(smiles):
    """Evaluate the chemical validity of a single molecule in terms of SMILES string

    Args:
      smiles: str, SMILES string.

    Returns:
      Boolean: if the SMILES string is a valid molecule

    """
    if smiles.strip() == "":
        return False
    mol = Chem.MolFromSmiles(smiles)
    if mol is None or mol.GetNumAtoms() == 0:
        return False
    return True


def validity(list_of_smiles):
    valid_list_smiles = list(filter(single_molecule_validity, list_of_smiles))
    return 1.0 * len(valid_list_smiles) / len(list_of_smiles)


def canonicalize(smiles):
    """Convert SMILES into canonical form.

    Args:
      smiles: str, SMILES string

    Returns:
      smiles: str, canonical SMILES string.

    """
    mol = Chem.MolFromSmiles(smiles)
    if mol is not None:
        return Chem.MolToSmiles(mol, isomericSmiles=True)
    else:
        return None


def unique_lst_of_smiles(list_of_smiles):
    canonical_smiles_lst = list(map(canonicalize, list_of_smiles))
    canonical_smiles_lst = list(
        filter(lambda x: x is not None, canonical_smiles_lst))
    canonical_smiles_lst = list(set(canonical_smiles_lst))
    return canonical_smiles_lst


def uniqueness(list_of_smiles):
    """Evaluate the uniqueness of a list of SMILES string, i.e., the fraction of unique molecules among a given list.

    Args:
      list_of_smiles: list (of SMILES string)

    Returns:
      uniqueness: float
    """
    canonical_smiles_lst = unique_lst_of_smiles(list_of_smiles)
    return 1.0 * len(canonical_smiles_lst) / len(list_of_smiles)


def novelty(generated_smiles_lst, training_smiles_lst):
    """Evaluate the novelty of set of generated smiles using list of training smiles as reference.
    Novelty is defined as the fraction of generated molecules that doesn't appear in the training set.

    Args:
      generated_smiles_lst: list (of SMILES string), which are generated.
      training_smiles_lst: list (of SMILES string), which are used for training.

    Returns:
      novelty: float
    """
    generated_smiles_lst = unique_lst_of_smiles(generated_smiles_lst)
    training_smiles_lst = unique_lst_of_smiles(training_smiles_lst)
    novel_ratio = (sum(
        [1 if i in training_smiles_lst else 0 for i in generated_smiles_lst]) *
                   1.0 / len(generated_smiles_lst))
    return 1 - novel_ratio


def diversity(list_of_smiles):
    """Evaluate the internal diversity of a set of molecules. The internbal diversity is defined as the average pairwise
      Tanimoto distance between the Morgan fingerprints.

    Args:
      list_of_smiles: list of SMILES strings

    Returns:
      div: float
    """
    list_of_unique_smiles = unique_lst_of_smiles(list_of_smiles)
    list_of_mol = [
        Chem.MolFromSmiles(smiles) for smiles in list_of_unique_smiles
    ]
    list_of_fp = [
        AllChem.GetMorganFingerprintAsBitVect(mol,
                                              2,
                                              nBits=2048,
                                              useChirality=False)
        for mol in list_of_mol
    ]
    avg_lst = []
    for idx, fp in enumerate(list_of_fp):
        for fp2 in list_of_fp[idx + 1:]:
            sim = DataStructs.TanimotoSimilarity(fp, fp2)
            ### option I
            distance = 1 - sim
            ### option II
            # distance = -np.log2(sim)
            avg_lst.append(distance)
    div = np.mean(avg_lst)
    return div


########################################
######## KL divergence ########


def _calculate_pc_descriptors(smiles, pc_descriptors):
    """Calculate Physical Chemical descriptors of a single SMILES (internal function).

    Args:
      list_of_smiles: SMILES strings
      pc_descriptors: list of strings, names of descriptors to calculate

    Returns:
      descriptros: list of float
    """
    from rdkit.ML.Descriptors import MoleculeDescriptors

    calc = MoleculeDescriptors.MolecularDescriptorCalculator(pc_descriptors)

    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return None
    _fp = calc.CalcDescriptors(mol)
    _fp = np.array(_fp)
    mask = np.isfinite(_fp)
    if (mask == 0).sum() > 0:
        logger.warning(f"{smiles} contains an NAN physchem descriptor")
        _fp[~mask] = 0

    return _fp


def calculate_pc_descriptors(smiles, pc_descriptors):
    """Calculate Physical Chemical descriptors of a list of molecules.

    Args:
      list_of_smiles: list of SMILES strings
      pc_descriptors: list of strings, names of descriptors to calculate

    Returns:
      descriptros: list of float
    """
    output = []

    for i in smiles:
        d = _calculate_pc_descriptors(i, pc_descriptors)
        if d is not None:
            output.append(d)

    return np.array(output)


def continuous_kldiv(X_baseline: np.array, X_sampled: np.array) -> float:
    """calculate KL divergence for two numpy arrays, conitnuous version.

    Args:
      X_baseline: numpy array
      X_sampled: numpy array

    Returns:
      KL divergence: float
    """
    X_baseline += 1e-5
    X_sampled += 1e-5
    from scipy.stats import entropy, gaussian_kde

    kde_P = gaussian_kde(X_baseline)
    kde_Q = gaussian_kde(X_sampled)
    x_eval = np.linspace(
        np.hstack([X_baseline, X_sampled]).min(),
        np.hstack([X_baseline, X_sampled]).max(),
        num=1000,
    )
    P = kde_P(x_eval) + 1e-10
    Q = kde_Q(x_eval) + 1e-10

    return entropy(P, Q)


def discrete_kldiv(X_baseline: np.array, X_sampled: np.array) -> float:
    """calculate KL divergence for two numpy arrays, discrete version.

    Args:
      X_baseline: numpy array
      X_sampled: numpy array

    Returns:
      KL divergence: float
    """
    from scipy.stats import entropy
    from scipy import histogram

    P, bins = histogram(X_baseline, bins=10, density=True)
    P += 1e-10
    Q, _ = histogram(X_sampled, bins=bins, density=True)
    Q += 1e-10

    return entropy(P, Q)


def get_fingerprints(mols, radius=2, length=4096):
    """
    Converts molecules to ECFP bitvectors.

    Args:
        mols: RDKit molecules
        radius: ECFP fingerprint radius
        length: number of bits

    Returns: a list of fingerprints
    """
    return [
        AllChem.GetMorganFingerprintAsBitVect(m, radius, length) for m in mols
    ]


def get_mols(smiles_list):
    """Convert SMILES strings to RDKit RDMol objects.

    Args:
      list_of_smiles: list of SMILES strings

    Returns:
      mols: list of RDKit RDMol objects
    """
    for i in smiles_list:
        try:
            mol = Chem.MolFromSmiles(i)
            if mol is not None:
                yield mol
        except Exception as e:
            logger.warning(e)


def calculate_internal_pairwise_similarities(smiles_list):
    """
    Computes the pairwise similarities of the provided list of smiles against itself.

    Args:
        smiles_list: list of str

    Returns:
        Symmetric matrix of pairwise similarities. Diagonal is set to zero.
    """
    if len(smiles_list) > 10000:
        logger.warning(f"Calculating internal similarity on large set of "
                       f"SMILES strings ({len(smiles_list)})")

    mols = get_mols(smiles_list)
    fps = get_fingerprints(mols)
    nfps = len(fps)

    similarities = np.zeros((nfps, nfps))

    for i in range(1, nfps):
        sims = DataStructs.BulkTanimotoSimilarity(fps[i], fps[:i])
        similarities[i, :i] = sims
        similarities[:i, i] = sims

    return similarities


def kl_divergence(generated_smiles_lst, training_smiles_lst):
    """Evaluate the KL divergence of set of generated smiles using list of training smiles as reference.
    KL divergence is defined as the averaged KL divergence of a set of physical chemical descriptors
    between a set of generated molecules and a set of training molecules.

    Args:
      generated_smiles_lst: list (of SMILES string), which are generated.
      training_smiles_lst: list (of SMILES string), which are used for training.

    Returns:
      KL divergence: float
    """
    pc_descriptor_subset = [
        "BertzCT",
        "MolLogP",
        "MolWt",
        "TPSA",
        "NumHAcceptors",
        "NumHDonors",
        "NumRotatableBonds",
        "NumAliphaticRings",
        "NumAromaticRings",
    ]

    def canonical(smiles):
        mol = Chem.MolFromSmiles(smiles)
        if mol is not None:
            return Chem.MolToSmiles(mol,
                                    isomericSmiles=True)  ### todo double check
        else:
            return None

    generated_lst_mol = list(map(canonical, generated_smiles_lst))
    training_lst_mol = list(map(canonical, training_smiles_lst))
    filter_out_func = lambda x: x is not None
    generated_lst_mol = list(filter(filter_out_func, generated_lst_mol))
    training_lst_mol = list(filter(filter_out_func, training_lst_mol))

    d_sampled = calculate_pc_descriptors(generated_lst_mol,
                                         pc_descriptor_subset)
    d_chembl = calculate_pc_descriptors(training_lst_mol, pc_descriptor_subset)

    kldivs = {}
    for i in range(4):
        kldiv = continuous_kldiv(X_baseline=d_chembl[:, i],
                                 X_sampled=d_sampled[:, i])
        kldivs[pc_descriptor_subset[i]] = kldiv

    # ... and for the int valued ones.
    for i in range(4, 9):
        kldiv = discrete_kldiv(X_baseline=d_chembl[:, i],
                               X_sampled=d_sampled[:, i])
        kldivs[pc_descriptor_subset[i]] = kldiv

    # pairwise similarity

    chembl_sim = calculate_internal_pairwise_similarities(training_lst_mol)
    chembl_sim = chembl_sim.max(axis=1)

    sampled_sim = calculate_internal_pairwise_similarities(generated_lst_mol)
    sampled_sim = sampled_sim.max(axis=1)

    kldiv_int_int = continuous_kldiv(X_baseline=chembl_sim,
                                     X_sampled=sampled_sim)
    kldivs["internal_similarity"] = kldiv_int_int
    """
        # for some reason, this runs into problems when both sets are identical.
        # cross_set_sim = calculate_pairwise_similarities(self.training_set_molecules, unique_molecules)
        # cross_set_sim = cross_set_sim.max(axis=1)
        #
        # kldiv_ext = discrete_kldiv(chembl_sim, cross_set_sim)
        # kldivs['external_similarity'] = kldiv_ext
        # kldiv_sum += kldiv_ext
  """

    # Each KL divergence value is transformed to be in [0, 1].
    # Then their average delivers the final score.
    partial_scores = [np.exp(-score) for score in kldivs.values()]
    score = sum(partial_scores) / len(partial_scores)
    return score


def fcd_distance_tf(generated_smiles_lst, training_smiles_lst):
    """Evaluate FCD distance between generated smiles set and training smiles set using tensorflow.

    Args:
      generated_smiles_lst: list (of SMILES string), which are generated.
      training_smiles_lst: list (of SMILES string), which are used for training.

    Returns:
      fcd_distance: float
    """
    import pkgutil, tempfile, os

    if "chemnet" not in globals().keys():
        global chemnet
        ### _load_chemnet
        chemnet_model_filename = "ChemNet_v0.13_pretrained.h5"
        model_bytes = pkgutil.get_data("fcd", chemnet_model_filename)
        tmpdir = tempfile.gettempdir()
        model_path = os.path.join(tmpdir, chemnet_model_filename)
        with open(model_path, "wb") as f:
            f.write(model_bytes)
        chemnet = fcd.load_ref_model(model_path)
        # _load_chemnet

    def _calculate_distribution_statistics(chemnet, molecules):
        sample_std = fcd.canonical_smiles(molecules)
        gen_mol_act = fcd.get_predictions(chemnet, sample_std)

        mu = np.mean(gen_mol_act, axis=0)
        cov = np.cov(gen_mol_act.T)
        return mu, cov

    mu_ref, cov_ref = _calculate_distribution_statistics(
        chemnet, training_smiles_lst)
    mu, cov = _calculate_distribution_statistics(chemnet, generated_smiles_lst)

    FCD = fcd.calculate_frechet_distance(mu1=mu_ref,
                                         mu2=mu,
                                         sigma1=cov_ref,
                                         sigma2=cov)
    fcd_distance = np.exp(-0.2 * FCD)
    return fcd_distance


def fcd_distance_torch(generated_smiles_lst, training_smiles_lst):
    """Evaluate FCD distance between generated smiles set and training smiles set using PyTorch.

    Args:
      generated_smiles_lst: list (of SMILES string), which are generated.
      training_smiles_lst: list (of SMILES string), which are used for training.

    Returns:
      fcd_distance: float
    """
    os.environ["KMP_DUPLICATE_LIB_OK"] = "True"
    from fcd_torch import FCD

    fcd = FCD(device="cpu", n_jobs=8)
    fcd_distance = fcd(generated_smiles_lst, training_smiles_lst)
    return fcd_distance


def fcd_distance(generated_smiles_lst, training_smiles_lst):
    """Evaluate FCD distance between generated smiles set and training smiles set.

    Args:
      generated_smiles_lst: list (of SMILES string), which are generated.
      training_smiles_lst: list (of SMILES string), which are used for training.

    Returns:
      fcd_distance: float
    """
    try:
        import tensorflow, fcd

        global fcd
    except:
        try:
            import torch, fcd_torch

            return fcd_distance_torch(generated_smiles_lst, training_smiles_lst)
        except:
            raise ImportError(
                "Please install fcd by 'pip install FCD' (for Tensorflow backend) \
                                            or 'pip install fcd_torch' (for PyTorch backend)!"
            )
    return fcd_distance_tf(generated_smiles_lst, training_smiles_lst)


def ncircles(list_of_smiles, t=0.75, random_state=42, radius=2, nbits=1024):
    '''
    NCircles is the maximum number of mutually exclusive circles that can fit into
    a set of molecular compounds with C of them as centres of circles of radius t.
    The set of molecular compounds is a chemical space measure, with the chemical
    space being all possible molecules and tanimoto distance the distance metric.

    This greedy algorithm approximates NCircles, as this is a known NP-hard problem.

    Input:
      list_of_smiles: List of chemical compounds in SMILES format
      t: hyperparameter controlling the radii of the circles
      random_state: random seed for reproducibility and debugging efficiency
      radius: radius used for converting to binary fingerprint vectors
      nbits: length of binary fingerprint vectors. Needed to compute distance.

    Output:
      k: NCircles score. This is a measure of diversity within a set of molecules
      C: binary fingerprint vectors of circle centres

    Reference:
    Xie Y, Xu Z, Ma J, et al. How Much Space Has Been Explored? Measuring the
    Chemical Space Covered by Databases and Machine-Generated Molecules[C]
    //The Eleventh International Conference on Learning Representations.
    '''
    import random

    def _ncircles_helper(fps):
        rng = random.Random(random_state)
        ids = list(range(len(fps)))
        rng.shuffle(ids)

        C_fps = []

        for i in ids:
            x = fps[i]
            if not C_fps:
                C_fps.append(x)
                continue

            min_d = min(
                (1.0 - DataStructs.TanimotoSimilarity(x, y)) for y in C_fps)

            if min_d > t:
                C_fps.append(x)

        return len(C_fps), C_fps

    if list_of_smiles and isinstance(list_of_smiles[0], str):
        fingerprints = [
            AllChem.GetMorganFingerprintAsBitVect(Chem.MolFromSmiles(s),
                                                  radius=radius,
                                                  nBits=nbits)
            for s in list_of_smiles
        ]
    else:
        fingerprints = list_of_smiles

    return _ncircles_helper(fingerprints)


def ncircles_recursive(list_of_smiles,
                       L,
                       m,
                       t=0.75,
                       random_state=42,
                       radius=2,
                       nbits=1024):
    '''
    See above for NCircles info. This is an adaptation of algorithm 4 in [1]. Recursively
    approximates NCircles of smaller subsets and combines for better results.

    Input:
      list_of_smiles: list of chemical compounds in SMILES format
      L: hyperparameter controlling the number of recursive layers
      m: hyperparameter controlling the max number of elements in a subset
      t: circle radii
      random state: random seed for reproducibility and debugging
      radius: radius for binary fingerprint vector
      nbits: length of binary fingerprint vector

    Output:
      k: NCircle score. Measure of diversity in chemical dataset
      C: binary fingerprint vectors of circle centres

    References:
    [1] Xie Y, Xu Z, Ma J, et al. How Much Space Has Been Explored? Measuring
    the Chemical Space Covered by Databases and Machine-Generated Molecules[C]
    //The Eleventh International Conference on Learning Representations.
    '''
    import random

    def _ncircles_rec_helper(fps, L):
        if L <= 0 or len(fps) == 0:
            return ncircles(fps, t, random_state, radius, nbits)

        rng = random.Random(random_state)
        ids = list(range(len(fps)))
        rng.shuffle(ids)

        # split into m approximately equal subsets
        subsets = [[] for _ in range(m)]
        for i, id in enumerate(ids):
            subsets[i % m].append(fps[id])

        all_centers = []

        for sub in subsets:
            if not sub:
                continue
            child_seed = None if random_state is None else rng.randrange(2**32)
            _, C_j = _ncircles_rec_helper(sub, L - 1)
            all_centers.extend(C_j)

        return ncircles(all_centers, t, random_state, radius, nbits)

    fingerprints = [
        AllChem.GetMorganFingerprintAsBitVect(Chem.MolFromSmiles(s),
                                              radius=radius,
                                              nBits=nbits)
        for s in list_of_smiles
    ]

    return _ncircles_rec_helper(fingerprints, L)


def hamiltonian_diversity(smiles=None,
                          mols=None,
                          dists=None,
                          radius=2,
                          nbits=1024):
    '''
    Hamiltonian Diversity is a molecular diversity metric inspired by the
    traveling salesman problem. It measures how diverse a given set of
    molecules is by computing the cost of the shortest Hamiltonian
    circuit through a graph of the chemical space measure weighted by
    its distance metric (here we use Tanimoto Distance on binary
    fingerprint vectors)

    Inputs:
      smiles: List of SMILES strings of molecules
      mols: Molecules from rdkit
      dists: pairwise distances in a molecular measure
      radius: radius for binary fingerprint vectors
      nbits: length of binary fingerprint vectors

    Outputs:
      HamDiv: hamiltonian diversity score (length of shortest
        hamiltonian path in chemical measure)

    Refencerences:
    Hu X, Liu G, Yao Q, et al. Hamiltonian diversity: effectively
    measuring molecular diversity by shortest Hamiltonian circuits[J].
    Journal of Cheminformatics, 2024, 16(1): 94.
    '''
    import networkx as nx

    if dists is not None:
        l = dists.shape[0]
    elif mols is not None:
        l = len(mols)
    elif smiles is not None:
        l = len(smiles)
        mols = [Chem.MolFromSmiles(s) for s in smiles]
    else:
        raise ValueError(
            "One of SMILES, Chem.Mol, or distances must be provided")

    if l <= 1:
        return 0.0

    # compute distance matrix if needed
    if dists is None:
        fps = [
            AllChem.GetMorganFingerprintAsBitVect(mol,
                                                  radius=radius,
                                                  nBits=nbits) for mol in mols
        ]
        dists = np.zeros((l, l))
        for i in range(l):
            for j in range(i + 1, l):
                if i == j:
                    dists[i, j] = 0.0
                    dists[j, i] = 0.0
                    continue

                dists[i,
                      j] = 1.0 - DataStructs.TanimotoSimilarity(fps[i], fps[j])
                dists[j, i] = dists[i, j]

    # construct graph
    G = nx.Graph()
    for i in range(l):
        for j in range(i + 1, l):
            G.add_edge(i, j, weight=dists[i, j])

    # solve TSP using greedy approach
    tsp = nx.approximation.greedy_tsp(G, weight='weight')

    return sum(dists[tsp[i - 1], tsp[i]] for i in range(1, len(tsp)))


def mol_vendi(smiles, radius=2, nbits=1024):
    '''
    Vendi Score is a measure of diversity in a molecular dataset.
    It is defined as the exponential of the Shannon entropy of the
    eigenvalues of the similarity matrix of a dataset. For more
    info refer to the source paper.

    Input:
      smiles: list of molecules in SMILES format
      radius: to be used to get binary fingerprint vectors
      nbits: length of binary fingerprint vectors

    Outputs:
      vendi score: measure of diversity within the chemical space

    References:
      Friedman, D., & Dieng, A. B. (2022). The vendi score: A
      diversity evaluation metric for machine learning. arXiv
      preprint arXiv:2210.02410.
    '''
    try:
        from vendi_score import vendi

        # get binary fingerprint vectors
        fingerprints = [
            AllChem.GetMorganFingerprintAsBitVect(Chem.MolFromSmiles(s),
                                                  radius=radius,
                                                  nBits=nbits) for s in smiles
        ]

        # construct similarity matrix K with Fingerprint Similarity
        N = len(smiles)
        K = np.zeros((N, N))

        for i in range(N):
            for j in range(i, N):
                K[i, j] = DataStructs.FingerprintSimilarity(
                    fingerprints[i], fingerprints[j])
                K[j, i] = K[i, j]

        return vendi.score_K(K)
    except:
        raise ImportError('Please install vendi_score to access vendi scores.')


def load_posecheck():
    '''
    Loads the PoseCheck class, which measures feasibility of a chemical
    configuration suggested by a generative model. For more info refer
    to the source paper.

    Outputs:
      PoseCheck: PoseCheck class, which acts as an API interface to
      get PoseCheck scores

    References:
      Harris C, Didi K, Jamasb A, et al. PoseCheck: Generative Models for
      3D Structure-based Drug Design Produce Unrealistic Poses[C]//NeurIPS
      2023 Generative AI and Biology (GenBio) Workshop.
    '''
    try:
        from posecheck import PoseCheck

        pc = PoseCheck()

        return pc
    except:
        raise ImportError('Please install posecheck to access PoseCheck')
